import functools
import sqlite3

try:
    from contextlib import nullcontext
except ImportError:
    from contextlib import suppress as nullcontext

import pytest
from RPA.Database import Database

from . import RESOURCES_DIR, RESULTS_DIR, temp_filename


DB_PATH = str(RESULTS_DIR / "database.db")
RETURNING_REASON = "On some systems SQLite fails to recognize the RETURNING properly"


def _get_library(**kwargs):
    lib = Database()
    lib.connect_to_database("sqlite3", DB_PATH, **kwargs)
    return lib


@pytest.fixture
def library():
    return _get_library(autocommit=True)


@pytest.fixture
def library_no_commit():
    return _get_library(autocommit=False)


def _ensure_orders(query):
    query("DROP TABLE IF EXISTS orders;")
    query("CREATE TABLE orders(id INTEGER PRIMARY KEY, name TEXT);")
    query('INSERT INTO orders(id, name) VALUES(1, "my-1st-order"),(2, "my-2nd-order");')


@pytest.mark.xfail(reason=RETURNING_REASON)
def test_query(library):
    _ensure_orders(library.query)
    orders_ids = library.query(
        'INSERT INTO orders(id, name) VALUES(3, "my-3rd-order");'
    )
    assert orders_ids[0][0] == 3

    orders = library.query("SELECT * FROM orders")
    order_names = [order["name"] for order in orders]
    assert order_names == ["my-1st-order", "my-2nd-order", "my-3rd-order"]


@pytest.mark.xfail(reason=RETURNING_REASON)
def test_query_parameterized_data(library):
    _ensure_orders(library.query)
    # This test uses a hardcoded string for simplicity, but in a real scenario the untrusted data
    # would be coming from an external source (Excel file, HTTP request, etc.)
    untrusted_data = "my-3rd-order"

    # Each database uses a particular style of formatting. SQLite3 uses '?' for instance.
    # For additional information about specific databases see https://bobby-tables.com/python
    orders_ids = library.query(
        "INSERT INTO orders(id, name) VALUES(3, ?);", data=(untrusted_data,)
    )

    assert orders_ids[0][0] == 3

    orders = library.query("SELECT * FROM orders")
    order_names = [order["name"] for order in orders]
    assert order_names == ["my-1st-order", "my-2nd-order", "my-3rd-order"]


@pytest.mark.parametrize("commit", [True, False])
def test_query_no_transaction(library_no_commit, commit):
    query = functools.partial(library_no_commit.query, sanstran=True)
    _ensure_orders(query)
    # No explicitly committing will loose all pushed records since auto-commit was
    # disabled in both the DB connection and `query` helper.
    if commit:
        query("COMMIT;")
    library_no_commit.disconnect_from_database()

    # No orders should be available in the database if they weren't committed.
    library_no_commit.connect_to_database("sqlite3", DB_PATH, autocommit=False)
    orders = query("SELECT * FROM orders", as_table=False)
    orders_length = 2 if commit else 0
    assert len(orders) == orders_length


@pytest.mark.parametrize(
    "should_fail, no_trans, orders_nr",
    [
        # Nothing fails since the script alone executes just fine.
        (False, False, 2),
        (False, True, 2),
        # It fails due to lastly added invalid statement.
        (True, False, 0),  # transaction enabled, a rollback is issued when it errors
        (True, True, 2),  # transaction disabled, no implicit rollback happens
    ],
)
def test_execute_sql_script(library_no_commit, should_fail, no_trans, orders_nr):
    sql_data = (RESOURCES_DIR / "script.sql").read_text()
    if should_fail:
        sql_data += "\nINVALID STATEMENT;"
        effect = pytest.raises(sqlite3.OperationalError)
    else:
        effect = nullcontext()
    with temp_filename(content=sql_data, suffix=".sql", mode="w") as sql_script:
        with effect:
            library_no_commit.execute_sql_script(sql_script, sanstran=no_trans)
    if no_trans:
        # Commit what was executed from the script. (even partially when it errors)
        # This way we guarantee the rollback functionality during a script execution
        #  that threw an error.
        library_no_commit.query("COMMIT;")

    orders = library_no_commit.query("SELECT * FROM orders")
    assert len(orders) == orders_nr


@pytest.mark.xfail(reason=RETURNING_REASON)
def test_query_explicit_returning(library):
    _ensure_orders(library.query)
    cursor = library.query(
        'INSERT INTO orders(id, name) VALUES(3, "my-3rd-order") RETURNING id;',
        returning=False,
    )
    assert cursor.fetchall()[0][0] == 3



import argparse
import os
from pathlib import Path


class FilePreparer:
    def fp_touch_file(self, path):
        Path(path).touch()

    def fp_change_dir(self, path):
        os.chdir(r"%s" % path)

    def fp_make_dirs(self, path):
        os.makedirs(path, exist_ok=True)

    def fp_write_file(self, path, content):
        try:
            if isinstance(content, str):
                content = content.encode("utf-8")
            with open(path, "wb") as fd:
                fd.write(content)
        except FileExistsError:
            pass

    def fp_prepare(self):
        self.fp_make_dirs("subfolder/first")
        self.fp_make_dirs("subfolder/second")
        self.fp_make_dirs("another/first")
        self.fp_make_dirs("another/second")
        self.fp_make_dirs("empty")

        self.fp_touch_file("emptyfile")
        self.fp_touch_file("notemptyfile")
        self.fp_touch_file("sorted1.test")
        self.fp_touch_file("sorted2.test")
        self.fp_touch_file("sorted3.test")
        self.fp_touch_file("subfolder/first/stuff.ext")
        self.fp_touch_file("subfolder/second/stuff.ext")
        self.fp_touch_file("another/first/noext")
        self.fp_touch_file("another/second/one")
        self.fp_touch_file("another/second/two")
        self.fp_touch_file("another/second/three")
        self.fp_touch_file("another/second/.dotfile")

        self.fp_write_file("notemptyfile", "some content here")
        self.fp_write_file("somebytes", b"\x00\x66\x6f\x6f\x00")

    def prepare_files_for_tests(self, root=os.getcwd()):
        self.fp_make_dirs(root)
        self.fp_change_dir(root)
        self.fp_prepare()


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        "root", default=os.getcwd(), help="Root directory for mock files"
    )
    args = parser.parse_args()

    preparer = FilePreparer()
    preparer.fp_make_dirs(args.root)
    preparer.fp_change_dir(args.root)

    preparer.fp_prepare()



import base64
import binascii
import copy
import json
import os
import yaml
from pathlib import Path

import mock
import pytest
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import padding
from cryptography.hazmat.primitives.ciphers.aead import AESGCM
from cryptography.hazmat.primitives.serialization import load_der_public_key

from RPA.Robocorp.Vault import (
    Vault,
    Secret,
    RobocorpVault,
    FileSecrets,
    BaseSecretManager,
    RobocorpVaultError,
)

RESOURCES = Path(__file__).parent / ".." / "resources"


class InvalidBaseClass(object):
    def get_secret(self, secret_name):
        assert False, "Should not be called"


class MockAdapter(BaseSecretManager):
    args = None
    name = None
    value = None

    def __init__(self, *args, **kwargs):
        MockAdapter.args = (args, kwargs)

    def get_secret(self, secret_name):
        MockAdapter.name = secret_name
        return MockAdapter.value

    def set_secret(self, secret):
        MockAdapter.name = secret.name
        MockAdapter.value = dict(secret)


@pytest.fixture
def mock_env_default(monkeypatch):
    monkeypatch.delenv("RPA_SECRET_MANAGER", raising=False)


@pytest.fixture
def mock_env_vault(monkeypatch):
    monkeypatch.setenv("RC_API_SECRET_HOST", "mock-url")
    monkeypatch.setenv("RC_API_SECRET_TOKEN", "mock-token")
    monkeypatch.setenv("RC_WORKSPACE_ID", "mock-workspace")


@pytest.fixture(params=["secrets.json", "secrets.yaml"])
def secrets_file(request):
    return request.param


def test_secrets_vault_as_default(mock_env_default, mock_env_vault):
    library = Vault()
    assert isinstance(library.adapter, RobocorpVault)


def test_secrets_vault_missing_token(mock_env_default, mock_env_vault, monkeypatch):
    monkeypatch.delenv("RC_API_SECRET_TOKEN", raising=False)
    library = Vault()
    with pytest.raises(KeyError):
        _ = library.adapter


def test_secrets_custom_adapter_arguments(mock_env_default):
    library = Vault("pos-value", key="key-value", default_adapter=MockAdapter)
    library.get_secret("not-relevant")  # Adapter created on first request
    assert MockAdapter.args == (("pos-value",), {"key": "key-value"})


def test_secrets_custom_adapter_get_secret(mock_env_default):
    MockAdapter.value = "mock-secret"
    library = Vault(default_adapter=MockAdapter)
    assert library.get_secret("mock-name") == "mock-secret"
    assert MockAdapter.name == "mock-name"


def test_secrets_adapter_missing_import(monkeypatch):
    monkeypatch.setenv("RPA_SECRET_MANAGER", "RPA.AdapterNotExist")
    with pytest.raises(ValueError):
        Vault()


def test_secrets_adapter_invalid_baseclass(mock_env_default):
    with pytest.raises(ValueError):
        Vault(default_adapter=InvalidBaseClass)


def test_secret_properties():
    secret = Secret(
        name="name-value",
        description="description-value",
        values={},
    )

    assert secret.name == "name-value"
    assert secret.description == "description-value"


def test_secret_get():
    secret = Secret(
        name="name-value",
        description="description-value",
        values={"key_one": "value_one", "key_two": "value_two"},
    )

    assert secret["key_one"] == "value_one"
    assert secret["key_two"] == "value_two"
    with pytest.raises(KeyError):
        _ = secret["key_invalid"]


def test_secret_set():
    secret = Secret(
        name="name-value",
        description="description-value",
        values={"key_one": "value_one", "key_two": "value_two"},
    )

    secret["key_one"] = "one"
    secret["key_two"] = "two"

    assert secret["key_one"] == "one"
    assert secret["key_two"] == "two"
    with pytest.raises(KeyError):
        _ = secret["key_invalid"]


def test_secret_update():
    secret = Secret(
        name="name-value",
        description="description-value",
        values={"key_one": "value_one", "key_two": "value_two"},
    )

    secret.update({"key_three": "value_three"})
    expected = {
        "key_one": "value_one",
        "key_two": "value_two",
        "key_three": "value_three",
    }

    assert secret == expected


def test_secret_iterate():
    secret = Secret(
        name="name-value",
        description="description-value",
        values={"key_one": "value_one", "key_two": "value_two"},
    )

    assert list(secret) == ["key_one", "key_two"]


def test_secret_contains():
    secret = Secret(
        name="name-value",
        description="description-value",
        values={"key_one": "value_one", "key_two": "value_two"},
    )

    assert "key_two" in secret


def test_secret_print():
    secret = Secret(
        name="name-value",
        description="description-value",
        values={"key_one": "value_one", "key_two": "value_two"},
    )

    repr_string = repr(secret)
    assert "value_one" not in repr_string
    assert "value_two" not in repr_string

    str_string = str(secret)
    assert "value_one" not in str_string
    assert "value_two" not in str_string


def test_adapter_filesecrets_from_arg(monkeypatch, secrets_file):
    monkeypatch.delenv("RPA_SECRET_FILE", raising=False)

    adapter = FileSecrets(RESOURCES / secrets_file)
    secret = adapter.get_secret("windows")
    assert isinstance(secret, Secret)
    assert "password" in secret
    assert secret["password"] == "secret"


def test_adapter_filesecrets_from_env(monkeypatch, secrets_file):
    monkeypatch.setenv("RPA_SECRET_FILE", str(RESOURCES / secrets_file))

    adapter = FileSecrets()
    secret = adapter.get_secret("windows")
    assert isinstance(secret, Secret)
    assert "password" in secret
    assert secret["password"] == "secret"


def test_adapter_filesecrets_invalid_file(monkeypatch):
    monkeypatch.setenv("RPA_SECRET_FILE", str(RESOURCES / "not-a-file.json"))

    # Should not raise
    adapter = FileSecrets()
    assert adapter.data == {}


def test_adapter_filesecrets_invalid_file_extension(monkeypatch):
    # Note the invalid ".yamla" extension not recognized by the serializers.
    monkeypatch.setenv("RPA_SECRET_FILE", str(RESOURCES / "secrets.yamla"))

    # Should raise immediately, because this is invalid straight from the
    # configuration. Only certain extensions (formats) are accepted.
    with pytest.raises(ValueError):
        FileSecrets()


def test_adapter_filesecrets_unknown_secret(monkeypatch, secrets_file):
    monkeypatch.setenv("RPA_SECRET_FILE", str(RESOURCES / secrets_file))

    adapter = FileSecrets()
    with pytest.raises(KeyError):
        _ = adapter.get_secret("not-exist")


def test_adapter_filesecrets_saving(monkeypatch, tmp_path, secrets_file):
    tmp_file = tmp_path / secrets_file
    tmp_file.write_text((RESOURCES / secrets_file).read_text())
    monkeypatch.setenv("RPA_SECRET_FILE", str(tmp_file))

    adapter = FileSecrets()
    secret = adapter.get_secret("credentials")
    secret["sap"]["password"] = "my-different-secret"
    adapter.set_secret(secret)
    adapter.save()

    assert tmp_file.suffix in (".json", ".yaml")
    if tmp_file.suffix == ".json":
        loader = json.load
    else:
        loader = yaml.full_load
    secret_dict = loader(tmp_file.open())
    assert secret_dict["credentials"]["sap"]["password"] == "my-different-secret"


@mock.patch("RPA.Robocorp.Vault.requests")
def test_adapter_vault_request(mock_requests, mock_env_default, mock_env_vault):
    mock_requests.get.return_value.json.return_value = {
        "name": "mock-name",
        "description": "mock-desc",
        "value": {"mock-key": "mock-value"},
        "encryption": {},
    }

    def mock_decrypt(payload):
        # Decrypt tested separately
        payload = copy.deepcopy(payload)
        payload["values"] = payload.pop("value")
        return payload

    adapter = RobocorpVault()
    adapter._decrypt_payload = mock_decrypt

    secret = adapter.get_secret("mock-name")
    assert secret.name == "mock-name"
    assert secret.description == "mock-desc"
    assert secret["mock-key"] == "mock-value"

    mock_requests.get.assert_called_once_with(
        "mock-url/secrets-v1/workspaces/mock-workspace/secrets/mock-name",
        headers={"Authorization": "Bearer mock-token"},
        params={
            "encryptionScheme": "robocloud-vault-transit-v2",
            "publicKey": adapter._public_bytes,
        },
    )


@mock.patch("RPA.Robocorp.Vault.requests")
def test_adapter_vault_error(mock_requests, mock_env_vault):
    mock_requests.get.side_effect = RuntimeError("Some request error")

    adapter = RobocorpVault()
    with pytest.raises(RobocorpVaultError):
        adapter.get_secret("mock-name")


def test_adapter_vault_encryption(mock_env_vault):
    data = json.dumps({"mock-key": "mock-value"}).encode("utf-8")

    # Cloud encrypts secret with symmetric encryption
    key = AESGCM.generate_key(bit_length=128)
    aesgcm = AESGCM(key)
    nonce = os.urandom(12)
    ct_data = aesgcm.encrypt(binascii.hexlify(nonce), data, b"")

    # Cloud uses client-supplied public key to encrypt symmetric key
    adapter = RobocorpVault()
    public_key = load_der_public_key(
        base64.b64decode(adapter._public_bytes), default_backend()
    )

    ct_key = public_key.encrypt(
        key,
        padding.OAEP(
            mgf=padding.MGF1(algorithm=hashes.SHA256()),
            algorithm=hashes.SHA256(),
            label=None,
        ),
    )

    ct_data, auth_tag = ct_data[:-16], ct_data[-16:]

    # Binary fields are base64 encoded, except nonce/iv (will possibly change)
    payload = {
        "name": "mock-name",
        "description": "mock-desc",
        "value": base64.b64encode(ct_data).decode("utf-8"),
        "encryption": {
            "encryptionScheme": adapter.ENCRYPTION_SCHEME,
            "encryptedAES": base64.b64encode(ct_key).decode("utf-8"),
            "authTag": base64.b64encode(auth_tag).decode("utf-8"),
            "iv": base64.b64encode(nonce).decode("utf-8"),
        },
    }

    response = adapter._decrypt_payload(payload)
    assert response["name"] == "mock-name"
    assert response["values"]["mock-key"] == "mock-value"



import os
import tempfile
from collections import namedtuple, OrderedDict
from contextlib import contextmanager
from pathlib import Path

import pytest
from RPA.Tables import Table, Tables, Dialect


RESOURCES = Path(__file__).parent / ".." / "resources"

DATA_COLUMNS = ["one", "two", "three", "four"]

TUPLE_THREE = namedtuple("Three", ["one", "two", "three"])
TUPLE_FOUR = namedtuple("Four", ["one", "two", "three", "four"])
TUPLE_SPARSE = namedtuple("Sparse", ["one", "two", "four"])
TUPLE_EMPTY = namedtuple("Empty", [])

DATA_NAMEDTUPLE = [
    TUPLE_THREE(1, 2, 3),
    TUPLE_THREE("a", "b", "c"),
    TUPLE_SPARSE(1, 2, 4),
    TUPLE_EMPTY(),
    TUPLE_FOUR(1, 2, 3, 4),
    TUPLE_EMPTY(),
]

DATA_DICT_LIST = {
    "one": [1, "a", 1, None, 1, None],
    "two": [2, "b", 2, None, 2],
    "three": [3, "c", None, None, 3, None],
    "four": [None, None, 4, None, 4],
}

DATA_LIST_DICT = [
    {"one": 1, "two": 2, "three": 3},
    {"one": "a", "two": "b", "three": "c"},
    {"one": 1, "two": 2, "four": 4},
    {},
    {"one": 1, "two": 2, "three": 3, "four": 4},
    {},
]

DATA_LIST_LIST = [[1, 2, 3], ["a", "b", "c"], [1, 2, None, 4], [], [1, 2, 3, 4], []]

DATA_FIXTURE = {
    "dict-list": (DATA_DICT_LIST, None),
    "list-dict": (DATA_LIST_DICT, None),
    "list-list": (DATA_LIST_LIST, DATA_COLUMNS),
    "namedtuple": (DATA_NAMEDTUPLE, None),
}


@contextmanager
def temppath():
    with tempfile.NamedTemporaryFile() as fd:
        path = fd.name
    try:
        yield path
    finally:
        os.unlink(path)


@pytest.fixture
def library():
    return Tables()


@pytest.fixture(params=DATA_FIXTURE)
def table(request):
    data, columns = DATA_FIXTURE[request.param]
    return Table(data, columns)


def test_table_repr(table):
    assert str(table) == "Table(columns=['one', 'two', 'three', 'four'], rows=6)"


def test_table_compare(table):
    assert table == Table(DATA_NAMEDTUPLE)
    assert table != "not-comparable"


def test_table_from_table(table):
    copy = Table(table)
    assert copy.columns == table.columns
    assert copy.data == table.data

    copy = Table(table, columns=["first", "second", "third", "fourth"])
    assert copy.columns == ["first", "second", "third", "fourth"]
    assert copy.data == table.data


def test_table_from_dict():
    copy = Table(DATA_DICT_LIST)
    assert copy.columns == ["one", "two", "three", "four"]
    assert len(copy.data) == 6

    copy = Table(DATA_DICT_LIST, columns=["one", "two"])
    assert copy.columns == ["one", "two"]
    assert copy.data == Table(DATA_DICT_LIST).get(columns=["one", "two"]).data


def test_table_invalid_data():
    with pytest.raises(TypeError):
        Table("cool")


def test_table_columns(table):
    assert table.columns == ["one", "two", "three", "four"]


def test_table_index(table):
    assert table.index == [0, 1, 2, 3, 4, 5]


def test_table_pad_short(table):
    assert table[0] == [1, 2, 3, None]


def test_table_pad_sparse(table):
    assert table[2] == [1, 2, None, 4]


def test_table_empty_row(table):
    assert table[3] == [None, None, None, None]


def test_table_negative_row_index(table):
    assert table[-1] == [None, None, None, None]
    assert table[-2] == [1, 2, 3, 4]
    assert table[-3] == [None, None, None, None]


def test_table_negative_column_index(table):
    assert table[0, 1] == 2
    assert table[0, -1] is None
    assert table[0, -2] == 3


def test_table_slice_index(table):
    assert table[1:3] == [["a", "b", "c", None], [1, 2, None, 4]]


def test_table_length(table):
    assert len(table) == 6


def test_table_invalid_column(table):
    with pytest.raises(ValueError):
        table.get_column("not_exist")


def test_table_range_columns():
    table = Table(DATA_LIST_LIST)
    assert table.columns == [0, 1, 2, 3]


def test_table_named_columns():
    table = Table(DATA_NAMEDTUPLE, columns=["two", "four"])
    assert table.columns == ["two", "four"]
    assert table.index == [0, 1, 2, 3, 4, 5]
    assert table[0] == [2, None]
    assert table[4] == [2, 4]


def test_table_too_short_columns():
    with pytest.raises(ValueError):
        Table(DATA_LIST_LIST, columns=["two", "four"])


def test_table_duplicate_columns():
    with pytest.raises(ValueError):
        Table(DATA_NAMEDTUPLE, columns=["two", "four", "two"])


def test_table_iterate_tuples():
    table = Table(
        [{"valid_key": 1, "invalid-key1": 2, "invalid/key2": 3, "123invalidkey3": 4}]
    )

    assert table.columns == [
        "valid_key",
        "invalid-key1",
        "invalid/key2",
        "123invalidkey3",
    ]

    rows = list(table.iter_tuples(with_index=False))
    assert len(rows) == 1
    assert rows[0] == (1, 2, 3, 4)
    assert rows[0]._fields == (
        "valid_key",
        "invalid_key1",
        "invalid_key2",
        "invalidkey3",
    )


def test_table_iterate_tuples_invalid():
    table = Table([{"one": 1, "two": 2, "assert": 3, "": 4}])
    assert table.columns == [
        "one",
        "two",
        "assert",
        "",
    ]

    with pytest.raises(ValueError):
        list(table.iter_tuples(with_index=False))


@pytest.mark.parametrize(
    "data, columns", DATA_FIXTURE.values(), ids=DATA_FIXTURE.keys()
)
def test_keyword_create_table(data, columns, library):
    table = library.create_table(data)
    assert len(table) == 6


def test_keyword_export_table_as_list(library, table):
    exported = library.export_table(table)
    assert exported == [
        {"one": 1, "two": 2, "three": 3, "four": None},
        {"one": "a", "two": "b", "three": "c", "four": None},
        {"one": 1, "two": 2, "three": None, "four": 4},
        {"one": None, "two": None, "three": None, "four": None},
        {"one": 1, "two": 2, "three": 3, "four": 4},
        {"one": None, "two": None, "three": None, "four": None},
    ]


def test_keyword_export_table_as_dict(library, table):
    exported = library.export_table(table, with_index=True, as_list=False)
    assert exported == OrderedDict(
        {
            "index": [0, 1, 2, 3, 4, 5],
            "one": [1, "a", 1, None, 1, None],
            "two": [2, "b", 2, None, 2, None],
            "three": [3, "c", None, None, 3, None],
            "four": [None, None, 4, None, 4, None],
        }
    )


def test_keyword_copy_table(library, table):
    copied = library.copy_table(table)
    assert copied == table


def test_keyword_clear_table(library, table):
    library.clear_table(table)
    assert len(table) == 0
    assert len(table.index) == 0
    assert table.columns == DATA_COLUMNS


def test_merge_tables(library):
    prices = {"Name": ["Egg", "Cheese", "Ham"], "Price": [10.0, 15.0, 20.0]}
    stock = {"Name": ["Egg", "Cheese", "Ham", "Spider"], "Stock": [12, 99, 0, 1]}

    merged = library.merge_tables(Table(prices), Table(stock))
    assert len(merged) == 7
    assert merged.columns == ["Name", "Price", "Stock"]
    assert merged[None, "Name"] == [
        "Egg",
        "Cheese",
        "Ham",
        "Egg",
        "Cheese",
        "Ham",
        "Spider",
    ]

    merged = library.merge_tables(Table(prices), Table(stock), index="Name")
    assert len(merged) == 4
    assert merged.get_row(0) == {"Name": "Egg", "Price": 10.0, "Stock": 12}
    assert merged.get_row(1) == {"Name": "Cheese", "Price": 15.0, "Stock": 99}
    assert merged.get_row(2) == {"Name": "Ham", "Price": 20.0, "Stock": 0}
    assert merged.get_row(3) == {"Name": "Spider", "Price": None, "Stock": 1}


def test_keyword_get_table_dimensions(library, table):
    rows, columns = library.get_table_dimensions(table)
    assert rows == 6
    assert columns == 4


def test_keyword_rename_table_columns(library, table):
    library.rename_table_columns(table, ["a", "b", "c", "d"])
    assert table.columns == ["a", "b", "c", "d"]
    assert table.get_column("a", as_list=True) == [1, "a", 1, None, 1, None]

    library.rename_table_columns(table, ["1", None, "2"])
    assert table.columns == ["1", "b", "2", "d"]


def test_keyword_add_table_column(library, table):
    library.add_table_column(table, name="five")
    assert table.columns == ["one", "two", "three", "four", "five"]
    assert table[0] == [1, 2, 3, None, None]


def test_keyword_add_table_rows(library, table):
    library.add_table_row(table, ["x", "y", "z"])
    assert len(table) == 7
    assert table.index[-2] == 5
    assert table[-1] == ["x", "y", "z", None]


def test_keyword_add_table_rows_too_long(library, table):
    library.add_table_row(table, ["x", "y", "z", "i", "j", "k"])
    assert len(table) == 7
    assert table.index[-2] == 5
    assert table[-1] == ["x", "y", "z", "i"]


def test_keyword_get_table_row(library, table):
    assert library.get_table_row(table, 0) == {
        "one": 1,
        "two": 2,
        "three": 3,
        "four": None,
    }


def test_keyword_get_table_column(library, table):
    assert library.get_table_column(table, 0) == [1, "a", 1, None, 1, None]


def test_keyword_set_table_row(library, table):
    assert table[1] == ["a", "b", "c", None]
    library.set_table_row(table, 1, ["w", "x", "y", "z"])
    assert table[1] == ["w", "x", "y", "z"]


def test_keyword_set_table_column(library, table):
    library.set_table_column(table, "one", "NaN")
    for row in table:
        assert row["one"] == "NaN"


def test_keyword_pop_table_row(library, table):
    assert len(table) == 6
    assert table[0] == [1, 2, 3, None]

    row = library.pop_table_row(table, row=0, as_list=True)

    assert len(table) == 5
    assert table[0] == ["a", "b", "c", None]
    assert row == [1, 2, 3, None]


def test_keyword_pop_table_column(library, table):
    library.pop_table_column(table, "two")
    assert table.columns == ["one", "three", "four"]
    assert len(table) == 6
    assert table[0] == [1, 3, None]


def test_keyword_get_table_slice(library, table):
    result = library.get_table_slice(table)
    assert result == table

    result = library.get_table_slice(table, start=3)
    assert len(result) == 3

    result = library.get_table_slice(table, end=2)
    assert len(result) == 2

    result = library.get_table_slice(table, end=-1)
    assert len(result) == 5

    result = library.get_table_slice(table, start=2, end=3)
    assert len(result) == 1

    result = library.get_table_slice(table, start=3, end=2)
    assert len(result) == 0


def test_keyword_find_table_rows(library, table):
    matches = library.find_table_rows(table, "three", "==", 3)
    assert len(matches) == 2

    matches = library.find_table_rows(table, "four", "is", None)
    assert len(matches) == 4


def test_keyword_set_row_as_column_names(library, table):
    assert table.columns == ["one", "two", "three", "four"]
    assert len(table) == 6

    library.set_row_as_column_names(table, 4)
    assert table.columns == [1, 2, 3, 4]
    assert len(table) == 5


def test_keyword_table_head(library, table):
    head = library.table_head(table, count=3)
    assert isinstance(head, Table)
    assert len(head) == 3
    assert head[0] == table[0]
    assert head[-1] == table[2]


def test_keyword_table_head_list(library, table):
    head = library.table_head(table, count=3, as_list=True)
    assert isinstance(head, list)
    assert len(head) == 3
    assert head[0] == table[0]
    assert head[-1] == table[2]


def test_keyword_table_tail(library, table):
    tail = library.table_tail(table, count=2)
    assert len(tail) == 2
    assert tail[-1] == table[-1]


def test_keyword_get_table_cell(library, table):
    assert library.get_table_cell(table, 0, 0) == 1
    assert library.get_table_cell(table, 2, 3) == 4


def test_keyword_set_table_cell_existing(library, table):
    library.set_table_cell(table, 0, 0, 123)
    assert table[0, 0] == 123
    library.set_table_cell(table, 1, "one", 321)
    assert table[1, 0] == 321


def test_keyword_set_table_cell_new(library, table):
    assert table.dimensions == (6, 4)
    library.set_table_cell(table, 9, 7, ">9000")
    assert table.dimensions == (10, 8)
    assert table[9, 7] == ">9000"


def test_keyword_sort_table_by_column(library, table):
    library.add_table_column(table, name="five", values=["bbb", 2, 3, 1, 3, "aaa"])

    library.sort_table_by_column(table, "five", ascending=True)
    assert library.get_table_column(table, "five") == [1, 2, 3, 3, "aaa", "bbb"]
    assert library.get_table_column(table, "one") == [None, "a", 1, 1, None, 1]

    library.sort_table_by_column(table, "five", ascending=False)
    assert library.get_table_column(table, "five") == ["bbb", "aaa", 3, 3, 2, 1]
    assert library.get_table_column(table, "one") == [1, None, 1, 1, "a", None]


def test_keyword_group_table_by_column(library, table):
    groups = library.group_table_by_column(table, "three")
    assert len(groups) == 3
    for group in groups:
        column = library.get_table_column(group, "three")
        assert len(set(column)) == 1


def test_keyword_filter_table_by_column(library, table):
    library.filter_table_by_column(table, "two", "==", 2)
    assert len(table) == 3
    assert all(row["two"] == 2 for row in table)


def test_keyword_filter_table_by_column_in(library, table):
    library.filter_table_by_column(table, "two", "in", ["b", None])
    assert len(table) == 3
    assert all(row["two"] != 2 for row in table)


def test_keyword_filter_table_by_column_not_contains(library):
    table = Table(
        [
            {"type": "Something", "value": 1},
            {"type": "Test", "value": 2},
            {"type": "Whatever", "value": 3},
            {"type": "Nothing", "value": 4},
        ]
    )
    library.filter_table_by_column(table, "type", "not contains", "thing")
    assert table.data == [["Test", 2], ["Whatever", 3]]


def test_keyword_filter_empty_rows(library, table):
    library.filter_empty_rows(table)
    assert len(table) == 4
    assert table[-1] == [1, 2, 3, 4]


def test_keyword_trim_empty_rows(library, table):
    library.trim_empty_rows(table)
    assert len(table) == 5
    assert table[-1] == [1, 2, 3, 4]
    assert table[-2] == [None, None, None, None]


def test_trim_column_names(library, table):
    library.rename_table_columns(table, ["a ", "b", "  c", " d "])
    assert table.columns == ["a ", "b", "  c", " d "]

    before = table.dimensions
    library.trim_column_names(table)
    after = table.dimensions

    assert before == after
    assert table.columns == ["a", "b", "c", "d"]


def test_keyword_read_table_from_csv(library):
    table = library.read_table_from_csv(RESOURCES / "easy.csv")
    assert len(table) == 3
    assert table.columns == ["first", "second", "third"]
    assert table[0] == ["1", "2", "3"]


def test_keyword_read_table_from_csv_no_header(library):
    table = library.read_table_from_csv(RESOURCES / "easy.csv", header=False)
    assert len(table) == 4
    assert table.columns == [0, 1, 2]
    assert table[0] == ["first", "second", "third"]


def test_keyword_read_table_from_csv_dialect_string(library):
    table = library.read_table_from_csv(
        RESOURCES / "hard.csv", dialect="excel", header=True
    )
    assert len(table) == 100


def test_keyword_read_table_from_csv_encoding(library):
    table = library.read_table_from_csv(RESOURCES / "easy.csv", encoding="utf-8")
    assert len(table) == 3
    assert table.columns == ["first", "second", "third"]
    assert table[0] == ["1", "2", "3"]


def test_keyword_read_table_from_csv_longer_lines(library):
    table = library.read_table_from_csv(
        RESOURCES / "big.csv", header=False, delimiters=";"
    )
    assert len(table) == 3
    assert len(table.columns) == 72
    table[-1][1] == "1121321715"


def test_keyword_read_table_from_csv_extra(library):
    table = library.read_table_from_csv(
        RESOURCES / "extra.csv", column_unknown="whoknows"
    )
    assert len(table) == 4
    assert table.columns == ["first", "second", "third", "whoknows"]
    assert table[0] == ["1", "2", "3", None]
    assert table[2] == ["7", "8", "9", ["11", "12"]]


def test_keyword_read_table_from_csv_manual(library):
    table = library.read_table_from_csv(
        RESOURCES / "hard.csv", dialect=Dialect.Excel, header=True
    )
    assert len(table) == 100
    assert table.columns == [
        "Region",
        "Country",
        "Item Type",
        "Sales Channel",
        "Order Priority",
        "Order Date",
        "Order ID",
        "Ship Date",
        "Units Sold",
        "Unit Price",
        "Unit Cost",
        "Total Revenue",
        "Total Cost",
        "Total Profit",
    ]
    assert table[-1] == [
        "Sub-Saharan Africa",
        "Mozambique",
        "Household",
        "Offline",
        "L",
        "2/10/2012",
        "665095412",
        "2/15/2012",
        "5367",
        "668.27",
        "502.54",
        "3586605.09",
        "2697132.18",
        "889472.91",
    ]


def test_keyword_write_table_to_csv(library, table):
    with temppath() as path:
        library.write_table_to_csv(table, path)
        with open(path) as fd:
            data = fd.readlines()

    assert len(data) == 7
    assert data[0] == "one,two,three,four\n"


def test_keyword_write_table_to_csv_encoding(library, table):
    with temppath() as path:
        library.write_table_to_csv(table, path, encoding="utf-8")
        with open(path) as fd:
            data = fd.readlines()

    assert len(data) == 7
    assert data[0] == "one,two,three,four\n"


def test_keyword_write_table_to_csv_columns_range(library):
    table = Table([[1, 2, 3], [4, 5, 6]])

    with temppath() as path:
        library.write_table_to_csv(table, path)
        with open(path) as fd:
            data = fd.readlines()

    assert len(data) == 3
    assert data[0] == "0,1,2\n"
    assert data[1] == "1,2,3\n"


def test_keyword_write_table_to_csv_delimiter(library, table):
    with temppath() as path:
        library.write_table_to_csv(table, path, encoding="utf-8", delimiter=";")
        with open(path) as fd:
            data = fd.readlines()

    assert len(data) == 7
    assert data[0] == "one;two;three;four\n"


def test_import_with_integer_keys():
    data = [
        {1: "Sub Total", 2: "$85.00 "},
        {1: "Tax", 2: "$8.50 "},
        {1: "Total", 2: "$93.50 "},
    ]

    table = Table(data)
    assert table.dimensions == (3, 3)
    assert table[0, 0] is None

    table = Table(data, columns=("Field", "Value"))
    assert table.dimensions == (3, 2)


def test_set_cell_empty_table():
    table = Table()
    table.set_cell(0, 0, "value")
    assert table.dimensions == (1, 1)
    assert table[0, 0] == "value"


def test_create_table_1d_dict():
    data = {"a": 1, "b": 2, "c": 3}
    table = Table(data)
    assert len(table) == 1
    assert table.columns == ["a", "b", "c"]


def test_create_table_1d_list():
    data = [1, 2, 3]
    table = Table(data)
    assert len(table) == 3


def test_columns_without_data():
    data = []
    columns = ["one", "two", "three"]
    table = Table(data, columns=columns)
    assert table.dimensions == (0, 3)


def test_data_with_nonetype():
    data = [
        {"one": 1, "two": 2},
        {"one": 1, "two": 2, None: 3},
        {"two": 2},
        {"four": 4, None: 3},
    ]

    table = Table(data)
    assert len(table) == 4
    assert table.columns == ["one", "two", 2, "four"]
    assert table.data == [
        [1, 2, None, None],
        [1, 2, 3, None],
        [None, 2, None, None],
        [None, None, 3, 4],
    ]


def test_table_append_row():
    data = {"a": [1], "b": [2], "c": [3]}
    table = Table(data)
    table.append_row()



# pylint: disable=c-extension-no-member
# pylint: disable=too-many-lines
# pylint: disable=possibly-used-before-assignment
# pylint: disable=consider-using-dict-items
# pylint: disable=used-before-assignment
# pylint: disable=consider-using-min-builtin
# pylint: disable=consider-using-max-builtin
from enum import Enum
import json
import logging
import os
import platform
import re
import subprocess
import time
from pathlib import Path
from typing import Any, Optional

from RPA.Desktop import Desktop
from RPA.Desktop.Clipboard import Clipboard
from RPA.Desktop.OperatingSystem import OperatingSystem
from RPA.core.geometry import Region
from RPA.core.helpers import delay, clean_filename
from RPA.core.locators import ImageLocator
from RPA.core.logger import deprecation


if platform.system() == "Windows":
    import ctypes
    import win32api
    import win32com.client
    import win32con
    import win32security
    import pywinauto
    import win32gui
    from comtypes import COMError
else:
    logging.getLogger(__name__).warning(
        "RPA.Desktop.Windows library works only on Windows platform"
    )


def write_element_info_as_json(
    elements: Any, filename: str, path: str = "output/json"
) -> None:
    """Write list of elements into json file

    :param elements: list of elements to write
    :param filename: output file name
    :param path: output directory, defaults to "output/json"
    """
    elements = elements if isinstance(elements, list) else [elements]
    filename = Path(f"{path}/{filename}.json")
    os.makedirs(filename.parent, exist_ok=True)
    with open(filename, "w", encoding="utf-8") as outfile:
        json.dump(elements, outfile, indent=4, sort_keys=True)


class ElementNotFoundError(Exception):
    """Raised when expected element is not found"""


class MenuItemNotFoundError(Exception):
    """Raised when expected menu item is not found"""


class UnknownWindowsBackendError(Exception):
    """Raised when unknown Windows backend is set"""


class ApplicationNotStarted(Exception):
    """Raised when application fails to start"""


SUPPORTED_BACKENDS = ["uia", "win32"]
WINDOWS_LOCATOR_STRATEGIES = {
    "name": "name",
    "class_name": "class_name",
    "class": "class_name",
    "control_type": "control_type",
    "type": "control_type",
    "automation_id": "automation_id",
    "id": "automation_id",
    "partial name": "partial name",
    "regexp": "regexp",
    "parent": "parent",
}


class Speed(Enum):
    """Values for pywinauto Timings"""

    DEFAULT = 1
    SLOW = 2
    FAST = 3


def to_speed(value):
    """Convert value to Speed enum."""
    if isinstance(value, Speed):
        return value

    sanitized = str(value).lower().strip().replace(" ", "_")
    try:
        return Speed[sanitized]
    except KeyError as err:
        raise ValueError(f"Unknown speed: {value}") from err


class Windows(OperatingSystem):
    """`Windows` is a library for managing the Windows operating system.

    **DEPRECATION WARNING! USE RPA.Windows library instead.**

    For Windows desktop automation Robocorp recommends the ``RPA.Windows`` library.

    No further updates will be released for this library and new functionality will continue
    to be developed in ``RPA.Windows`` library.

    **Running Windows applications**

    Windows applications can be started in several ways. The library supports
    the following keywords:

    - Open Application (dispatch Office applications)
    - Open File (open the file as process which opens the associated application)
    - Open Executable (uses pywinauto start)
    - Open Using Run Dialog (uses Windows run dialog)
    - Open From Search (uses Windows search dialog)

    **Locators**

    `Locator` is used to identify the element for interaction - usually for a mouse click.

    Locators can investigated for application once it has been opened by calling
    the keyword `get_windows_elements` which can store locator information into JSON file
    and `screenshot` of the element into an image file.

    **Identifying locator**

    The element needs to be identified by a unique method, for example, "Three" for button 3
    in the Calculator application. It can be given either as `Three` or `name:Three`.

    Possible search criterias:

    - name
    - class (class_name)
    - type (control_type)
    - id (automation_id)
    - any if none was defined

    The current method of inspecting elements on Windows is `inspect.exe` which is part
    of `Windows SDK <https://docs.microsoft.com/en-us/windows/win32/winauto/inspect-objects>`_.

    **Keyboard**

    The keyword `send_keys` can be used to send keys to the active window. The keyword
    `type_keys` sends keys to the active window element.

    Special key codes are documented on `pywinauto <https://pywinauto.readthedocs.io/en/latest/code/pywinauto.keyboard.html#>`_
    documentation page.

    **FAQ**

    Q. I see error message `AttributeError: module 'win32com.gen_py.00020813-0000-0000-C000-000000000046x0x1x9' has no attribute 'CLSIDToClassMap'`

    A. From PowerShell run this command: `Remove-Item -path $env:LOCALAPPDATA\\Temp\\gen_py -recurse`

    **Examples**

    **Robot Framework**

    .. code-block:: robotframework

        *** Settings ***
        Library          RPA.Desktop.Windows
        Suite Teardown   Close all applications

        *** Tasks ***
        Open Calculator using run dialog
            ${result}=              Open using run dialog    calc.exe   Calculator
            ${result}=              Get Window Elements
            Send Keys               5*2=
            ${result}=              Get element             partial name:Display is
            Log Many                ${result}
            ${result}=              Get element rich text   id:CalculatorResults
            Should Be Equal As Strings  ${result}  Display is 10
            ${result}=              Get element rectangle   partial name:Display is
            ${result}=              Is Element Visible      CalculatorResults
            ${result}=              Is Element Enabled      partial name:Display is

    **Python**

    .. code-block:: python

        from RPA.Desktop.Windows import Windows

        win = Windows()

        def open_calculator():
            win.open_from_search("calc.exe", "Calculator")
            elements = win.get_window_elements()

        def make_calculations(expression):
            win.send_keys(expression)
            result = win.get_element_rich_text('id:CalculatorResults')
            return int(result.strip('Display is '))

        if __name__ == "__main__":
            open_calculator()
            exp = '5*2='
            result = make_calculations(exp)
            print(f"Calculation result of '{exp}' is '{result}'")
            win.close_all_applications()
    """  # noqa: E501

    ROBOT_LIBRARY_SCOPE = "GLOBAL"
    ROBOT_LIBRARY_DOC_FORMAT = "REST"

    def __init__(self, backend: str = "uia") -> None:
        deprecation(
            "`RPA.Desktop.Windows` got deprecated and will be no longer maintained, "
            "please use `RPA.Windows` instead "
            "(https://robocorp.com/docs/libraries/rpa-framework/rpa-windows)"
        )

        OperatingSystem.__init__(self)
        self._apps = {}
        self._app_instance_id = 0
        self._active_app_instance = -1
        self.set_windows_backend(backend)
        self.app = None
        self.dlg = None
        self.windowtitle = None
        self.logger = logging.getLogger(__name__)
        self.clipboard = Clipboard()
        self.elements = None
        self.controls = None

    def set_windows_backend(self, backend: str) -> None:
        """Set Windows backend which is used to interact with Windows
        applications

        Allowed values defined by `SUPPORTED_BACKENDS`

        :param backend: name of the backend to use

        Example:

        .. code-block:: robotframework

            Set Windows Backend   uia
            Open Executable   calc.exe  Calculator
            Set Windows Backend   win32
            Open Executable   calc.exe  Calculator

        """
        if backend and backend.lower() in SUPPORTED_BACKENDS:
            self._backend = backend.lower()
        else:
            raise UnknownWindowsBackendError(
                "Unsupported Windows backend: %s" % backend
            )

    def _add_app_instance(
        self,
        app: Any = None,
        dialog: bool = True,
        params: dict = None,
    ) -> int:
        params = params or {}
        self._app_instance_id += 1
        process_id = None
        handle = None
        if app:
            self.app = app
            if hasattr(app, "process"):
                process_id = app.process
            handle = win32gui.GetForegroundWindow()

        default_params = {
            "app": app,
            "id": self._app_instance_id,
            "dialog": dialog,
            "process_id": process_id,
            "handle": handle,
            "dispatched": False,
        }

        self._apps[self._app_instance_id] = {**default_params, **params}

        self.logger.debug(
            "Added app instance %s: %s",
            self._app_instance_id,
            self._apps[self._app_instance_id],
        )
        self._active_app_instance = self._app_instance_id
        return self._active_app_instance

    def switch_to_application(self, app_id: int) -> None:
        """Switch to application by id.

        :param app_id: application's id
        :raises ValueError: if application is not found by given id

        Example:

        .. code-block:: robotframework

            ${app1}    Open Application   Excel
            ${app2}    Open Application   Word
            Switch To Application   ${app1}

        """
        if app_id and app_id in self._apps.keys():
            app = self.get_app(app_id)
            self._active_app_instance = app_id
            self.app = app["app"]
            if "windowtitle" in app:
                self.open_dialog(app["windowtitle"], existing_app=True)
                delay(0.5)
                self.restore_dialog(app["windowtitle"])
        else:
            raise ValueError(f"No open application with id '{app_id}'")

    def get_open_applications(self):
        """Get list of all open applications

        Returns a dictionary

        Example:

        .. code-block:: robotframework

            ${app1}    Open Application   Excel
            ${app2}    Open Executable    calc.exe  Calculator
            ${app3}    Open File          /path/to/myfile.txt
            &{apps}    Get Open Applications

        """
        return self._apps

    def get_app(self, app_id: int = None) -> Any:
        """Get application object by id

        By default returns active_application application object.

        :param app_id: id of the application to get, defaults to None
        :return: application object

        Example:

        .. code-block:: robotframework

            ${app1}        Open Application   Excel
            &{appdetails}  Get App   ${app1}

        """
        if app_id is None and self._active_app_instance != -1:
            return self._apps[self._active_app_instance]
        else:
            return self._apps[app_id]

    def open_application(self, application: str) -> int:
        """Open application by dispatch method

        This keyword is used to launch Microsoft applications like
        Excel, Word, Outlook and Powerpoint.

        :param application: name of the application as `str`
        :return: application instance id

        Example:

        .. code-block:: robotframework

            ${app1}    Open Application   Excel
            ${app2}    Open Application   Word

        """
        self.logger.info("Open application: %s", application)
        app = win32com.client.gencache.EnsureDispatch(f"{application}.Application")
        app.Visible = True
        # show eg. file overwrite warning or not
        if hasattr(self.app, "DisplayAlerts"):
            app.DisplayAlerts = False
        params = {
            "dispatched": True,
            "startkeyword": "Open Application",
        }
        return self._add_app_instance(app, dialog=False, params=params)

    def open_file(
        self,
        filename: str,
        windowtitle: str = None,
        wildcard: bool = False,
        timeout: int = 10,
    ) -> Optional[int]:
        """Open associated application when opening file

        Keyword `Open Dialog` is used if `windowtitle` is given.

        :param filename: path to file
        :param windowtitle: name of the window
        :param wildcard: set True for inclusive window title search, default False
        :param timeout: time to wait for dialog to appear
        :return: application id or None

        Example:

        .. code-block:: robotframework

            ${app1}    Open File   /path/to/myfile.txt

        """
        self.logger.info("Open file: %s", filename)
        if platform.system() == "Windows":
            # pylint: disable=no-member
            os.startfile(filename)
        elif platform.system() == "Darwin":
            subprocess.call(["open", filename])
        else:
            subprocess.call(["xdg-open", filename])

        app_instance = None
        if windowtitle:
            app_instance = self.open_dialog(
                windowtitle, wildcard=wildcard, timeout=timeout
            )
            if app_instance > 0:
                self._apps[app_instance]["executable"] = filename
                self._apps[app_instance]["startkeyword"] = "Open File"

        return app_instance

    def open_executable(
        self,
        executable: str,
        windowtitle: str,
        backend: str = None,
        work_dir: str = None,
        wildcard: bool = False,
    ) -> int:
        """Open Windows executable. Window title name is required
        to get handle on the application.

        :param executable: name of the executable
        :param windowtitle: name of the window
        :param backend: set Windows backend, default None means using
         library default value
        :param work_dir: path to working directory, default None
        :param wildcard: set True for inclusive window title search, default False
        :return: application instance id

        Example:

        .. code-block:: robotframework

            ${app1}    Open Executable   calc.exe  Calculator
            ${app2}    Open Executable   notepad.exe  Notepad   wildcard=True

        """
        self.logger.info("Opening executable: %s - window: %s", executable, windowtitle)
        if backend:
            self.set_windows_backend(backend)

        app = pywinauto.Application(backend=self._backend).start(
            cmd_line=executable, work_dir=work_dir
        )
        app_instance = self.open_dialog(windowtitle, wildcard=wildcard)
        self._apps[app_instance]["app"] = app
        self._apps[app_instance]["executable"] = executable
        self._apps[app_instance]["startkeyword"] = "Open Executable"
        return app_instance

    def open_using_run_dialog(
        self,
        executable: str,
        windowtitle: str,
        timeout: int = 10,
        wildcard: bool = False,
    ) -> int:
        """Open application using Windows run dialog.
        Window title name is required to get handle on the application.

        :param executable: name of the executable
        :param windowtitle: name of the window
        :param timeout: time to wait for dialog to appear
        :param wildcard: set True for inclusive window title search, default False
        :return: application instance id

        Example:

        .. code-block:: robotframework

            ${app1}    Open Using Run Dialog  notepad  Untitled - Notepad
            ${app2}    Open Using Run Dialog  notepad  Notepad    wildcard=True

        """
        self.send_keys("{VK_LWIN down}r{VK_LWIN up}")
        delay(1)

        self.send_keys_to_input(executable, send_delay=0.2, enter_delay=0.5)

        app_instance = self.open_dialog(windowtitle, timeout=timeout, wildcard=wildcard)
        self._apps[app_instance]["executable"] = executable
        self._apps[app_instance]["startkeyword"] = "Open Using Run Dialog"
        return app_instance

    def open_from_search(
        self,
        executable: str,
        windowtitle: str,
        timeout: int = 10,
        wildcard: bool = False,
    ) -> int:
        """Open application using Windows search dialog.
        Window title name is required to get handle on the application.

        :param executable: name of the executable
        :param windowtitle: name of the window
        :param timeout: time to wait for dialog to appear
        :param wildcard: set True for inclusive window title search, default False
        :return: application instance id

        Example:

        .. code-block:: robotframework

            ${app1}    Open From Search  calculator  Calculator
            ${app2}    Open From Search  notepad  Notepad  wildcard=True

        """
        self.logger.info("Run from start menu: %s", executable)
        self.send_keys("{LWIN}")
        delay(1)

        self.send_keys_to_input(executable)

        app_instance = None
        for _ in range(10):
            try:
                app_instance = self.open_dialog(
                    windowtitle, timeout=timeout, wildcard=wildcard
                )
            except AttributeError:
                pass
            except COMError:
                pass
            if app_instance:
                self._apps[app_instance]["executable"] = executable
                self._apps[app_instance]["startkeyword"] = "Open From Search"
                break
            else:
                time.sleep(0.5)
        if not app_instance:
            raise ApplicationNotStarted("Unable to get application instance")
        return app_instance

    def get_spaced_string(self, text):
        """Replace spaces in a text with `pywinauto.keyboard`
        space characters `{VK_SPACE}`

        :param text: replace spaces in this string

        Example:

        .. code-block:: robotframework

            ${txt}    Get Spaced String   My name is Bond
            # ${txt} = My{VK_SPACE}name{VK_SPACE}is{VK_SPACE}Bond
            Send Keys To Input  ${txt}

        """
        return text.replace(" ", "{VK_SPACE}")

    def send_keys_to_input(
        self,
        keys_to_type: str,
        with_enter: bool = True,
        send_delay: float = 0.5,
        enter_delay: float = 1.5,
    ) -> None:
        """Send keys to windows and add ENTER if `with_enter` is True

        At the end of send_keys there is by default 0.5 second delay.
        At the end of ENTER there is by default 1.5 second delay.

        :param keys_to_type: keys to type into Windows
        :param with_enter: send ENTER if `with_enter` is True
        :param send_delay: delay after send_keys
        :param enter_delay: delay after ENTER

        Example:

        .. code-block:: robotframework

            ${txt}    Get Spaced String   My name is Bond, James Bond
            Send Keys To Input  ${txt}    with_enter=False
            Send Keys To Input  {ENTER}THE   send_delay=5.0  with_enter=False
            Send Keys To Input  {VK_SPACE}-{VK_SPACE}END   enter_delay=5.0

        """
        # Set keyboard layout for Windows platform
        if platform.system() == "Windows":
            win32api.LoadKeyboardLayout("00000409", 1)

        self.send_keys(keys_to_type)
        delay(send_delay)
        if with_enter:
            self.send_keys("{ENTER}")
            delay(enter_delay)

    def minimize_dialog(self, windowtitle: str = None) -> None:
        """Minimize window by its title

        :param windowtitle: name of the window, default `None` means that
         active window is going to be minimized

        Example:

        .. code-block:: robotframework

            Open Using Run Dialog  calc     Calculator
            Open Using Run Dialog  notepad  Untitled - Notepad
            Minimize Dialog    # Current window (Notepad)
            Minimize Dialog    Calculator

        """
        windowtitle = (
            windowtitle or self._apps[self._active_app_instance]["windowtitle"]
        )
        self.logger.info("Minimize dialog: %s", windowtitle)
        self.dlg = pywinauto.Desktop(backend=self._backend)[windowtitle]
        self.dlg.minimize()

    def restore_dialog(self, windowtitle: str = None) -> None:  # noqa: C901
        """Restore window by its title

        :param windowtitle: name of the window, default `None` means that
         active window is going to be restored

        Example:

        .. code-block:: robotframework

            Open Using Run Dialog  notepad  Untitled - Notepad
            Minimize Dialog
            Sleep             1s
            Restore Dialog
            Sleep             1s
            Restore Dialog    Untitled - Notepad

        """
        # TODO. Handle too compled method
        self.logger.info("Restore dialog: %s", windowtitle)
        if windowtitle is None and self._active_app_instance == -1:
            raise ValueError(
                "There are no applications opened by library or window title is empty"
            )
        app = None
        handle = None
        if windowtitle is None:
            app = self._apps[self._active_app_instance]["app"]
            handle = self._apps[self._active_app_instance]["handle"]
            windowtitle = self._apps[self._active_app_instance]["windowtitle"]
        else:
            for app_id, app in self._apps.items():
                if "windowtitle" in app.keys() and windowtitle == app["windowtitle"]:
                    application_id = app_id
                    app = self._apps[application_id]["app"]
                    handle = self._apps[application_id]["handle"]
        try:
            if app and handle > 0:
                app.window(handle=handle).restore()
            else:
                wins = self.get_window_list()
                for win in wins:
                    if windowtitle == win["title"]:
                        handle = win["handle"]
                        break
                if handle and handle > 0:
                    self.connect_by_handle(handle, windowtitle=windowtitle)
                    self.restore_dialog()
                else:
                    raise ValueError("Could not restore dialog: %s" % windowtitle)
        except pywinauto.findwindows.ElementAmbiguousError as e:
            self.logger.info("Could not restore dialog, %s", str(e))

    def open_dialog(
        self,
        windowtitle: str = None,
        highlight: bool = False,
        timeout: int = 10,
        existing_app: bool = False,
        wildcard: bool = False,
        parse_elements: bool = True,
    ) -> Any:
        """Open window by its title.

        :param windowtitle: name of the window, defaults to active window if None
        :param highlight: draw outline for window if True, default False
        :param timeout: time to wait for dialog to appear
        :param existing_app: set True if selecting window which library has already
         accessed, default False
        :param wildcard: set True for inclusive window title search, default False
        :param parse_elements: set False to not to parse elements of the window,
         default True

        Example:

        .. code-block:: robotframework

            Open Dialog       Untitled - Notepad
            Open Dialog       Untitled - Notepad   highlight=True   timeout=5
            Open Dialog       Notepad   wildcard=True

        """
        self.logger.info("Open dialog: '%s'", windowtitle)

        app_instance = None
        end_time = time.time() + float(timeout)
        while time.time() < end_time and app_instance is None:
            for window in self.get_window_list():
                if (not wildcard and windowtitle == window["title"]) or (
                    wildcard and windowtitle in window["title"]
                ):
                    self.windowtitle = window["title"]
                    app_instance = self.connect_by_handle(
                        window["handle"],
                        windowtitle=self.windowtitle,
                        existing_app=existing_app,
                        parse_elements=parse_elements,
                    )
                    break
            time.sleep(0.1)

        if app_instance is None:
            raise ValueError(
                "No window with title '%s', wildcard: %s" % (windowtitle, wildcard)
            )

        if highlight:
            self.dlg.draw_outline()

        return app_instance

    def connect_by_pid(self, app_pid: str, windowtitle: str = None) -> Any:
        """Connect to application by its pid

        :param app_pid: process id of the application
        :param windowtitle: name of the window, defaults to active window if None

        Example:

        .. code-block:: robotframework

            ${appid}  Connect By PID  3231

        """
        self.logger.info("Connect to application pid: %s", app_pid)
        window_list = self.get_window_list()
        for win in window_list:
            if win["pid"] == app_pid:
                if windowtitle is None or (windowtitle and windowtitle in win["title"]):
                    self.logger.info(
                        "PID:%s matched window title:%s", win["pid"], win["title"]
                    )
                    return self.connect_by_handle(win["handle"], windowtitle)
        return None

    def connect_by_handle(
        self,
        handle: int,
        windowtitle: str = None,
        existing_app: bool = False,
        parse_elements: bool = True,
    ) -> Any:
        """Connect to application by its handle

        :param handle: handle of the application
        :param windowtitle: name of the window, defaults to active window if None
        :param existing_app: set True if selecting window which library has already
         accessed, default False
        :param parse_elements: set False to not to parse elements of the window,
         default True

        Example:

        .. code-block:: robotframework

            ${appid}  Connect By Handle  88112

        """
        self.logger.info("Connect to application handle: %s", handle)
        app_instance = None
        app = pywinauto.Application(backend=self._backend).connect(
            handle=handle, visible_only=False
        )
        self.dlg = app.window(handle=handle)
        self.dlg.set_focus()
        params = None
        if existing_app:
            for key in self._apps:
                if self._apps[key]["handle"] == handle:
                    app_instance = key
                    break
        else:
            if windowtitle is not None:
                params = {"windowtitle": windowtitle}
            app_instance = self._add_app_instance(app=app, params=params, dialog=False)
        if parse_elements:
            self.refresh_window()
        return app_instance

    def close_all_applications(self) -> None:
        """Close all applications

        Example:

        .. code-block:: robotframework

            Open Application   Excel
            Open Application   Word
            Open Executable    notepad.exe   Untitled - Notepad
            Close All Applications

        """
        self.logger.info("Closing all applications")
        self.logger.debug("Applications in memory: %d", len(self._apps))
        for aid in list(self._apps):
            self.quit_application(aid)
            del self._apps[aid]

    def quit_application(self, app_id: int = None, send_keys: bool = False) -> None:
        """Quit an application by application id or
        active application if `app_id` is None.

        :param app_id: application_id, defaults to None
        :param send_keys: if ALT+F4 should be used to quit, default False

        Example:

        .. code-block:: robotframework

            ${app1}   Open Application   Excel
            ${app2}   Open Application   Word
            Quit Application  ${app1}

        """
        app = self.get_app(app_id)
        self.logger.info("Quit application: %s (%s)", app_id, app)
        if send_keys:
            self.logger.info("Quit by F4 shortcut")
            self.switch_to_application(app_id)
            self.send_keys("%{F4}")
        else:
            if app["dispatched"]:
                self.logger.info("Quit by app.Quit()")
                app["app"].Quit()
            else:
                if "process_id" in app and app["process_id"] > 0:
                    # pylint: disable=E1101
                    pid = app["process_id"]
                    if self.process_id_exists(pid):
                        self.logger.info("Quit by killing process id")
                        self.kill_process_by_pid(pid)
                    else:
                        self.logger.info("Process pid '%s' did not exist anymore", pid)
                else:
                    self.logger.info("Quit by app.kill()")
                    app["app"].kill()
        self._active_app_instance = -1

    def type_keys(self, keys: str) -> None:
        """Type keys into active window element.

        :param keys: list of keys to type

        Example:

        .. code-block:: robotframework

            Open Executable  notepad.exe  Untitled - Notepad
            Type Keys   My text

        """
        if self.dlg is None:
            raise ValueError("No dialog open")
        self.dlg.type_keys(keys)

    def type_into(self, locator: str, keys: str, empty_field: bool = False) -> None:
        """Type keys into element matched by given locator.

        :param locator: element locator
        :param keys:    list of keys to type
        :param empty_field: if field should be emptied before typing, default False

        Example:

        .. code-block:: robotframework

            Open Executable  calc.exe  Calculator
            Type Into        CalculatorResults  11
            Type Into        CalculatorResults  22  empty_field=True

        """
        elements, _ = self.find_element(locator)
        if elements and len(elements) == 1:
            ctrl = elements[0]["control"]
            empty_method = "^a{BACKSPACE}" if empty_field else ""
            ctrl.type_keys(f"{empty_method}{keys}")
        else:
            raise ValueError(f"Could not find unique element for '{locator}'")

    def send_keys(self, keys: str) -> None:
        """Send keys into active windows.

        :param keys: list of keys to send

        Example:

        .. code-block:: robotframework

            Open Executable  calc.exe  Calculator
            Send Keys        2{+}3{=}

        """
        pywinauto.keyboard.send_keys(keys)

    def get_text(self, locator: str) -> dict:
        """Get text from element

        :param locator: element locator

        Example:

        .. code-block:: robotframework

            Open Using Run Dialog  calc     Calculator
            Type Into    CalculatorResults   11
            Type Into    CalculatorResults   55
            &{val}       Get Text   CalculatorResults

        """
        elements, _ = self.find_element(locator)
        element_text = {}
        if elements and len(elements) == 1:
            ctrl = elements[0]["control"]
            element_text["value"] = (
                str(ctrl.get_value()) if hasattr(ctrl, "get_value") else None
            )
            element_text["children_texts"] = (
                "".join(ctrl.children_texts())
                if hasattr(ctrl, "children_texts")
                else None
            )
            element_text["rich_text"] = (
                elements[0]["rich_text"] if "rich_text" in elements[0].keys() else ""
            )
            legacy = (
                ctrl.legacy_properties() if hasattr(ctrl, "legacy_properties") else None
            )
            element_text["legacy_value"] = str(legacy["Value"]) if legacy else None
            element_text["legacy_name"] = str(legacy["Name"]) if legacy else None
        return element_text

    def mouse_click(
        self,
        locator: str = None,
        x: int = 0,
        y: int = 0,
        off_x: int = 0,
        off_y: int = 0,
        image: str = None,
        method: str = "locator",
        ctype: str = "click",
        focus: str = "center",
        tolerance: Optional[int] = None,
    ) -> None:
        # pylint: disable=C0301
        """Mouse click `locator`, `coordinates`, or `image`

        When using method `locator`,`image` or `ocr` mouse is clicked by default at
        center coordinates.

        Click types are:

        - `click` normal left button mouse click
        - `double`
        - `right`

        :param locator: element locator on active window
        :param x: coordinate x on desktop
        :param y: coordinate y on desktop
        :param off_x: offset x (used for locator and image clicks)
        :param off_y: offset y (used for locator and image clicks)
        :param image: image to click on desktop
        :param method: one of the available methods to mouse click, default "locator"
        :param ctype: type of mouse click
        :param focus: default point for element click is 'center', can be set to 'topleft'
         to click top left corner of the element
        :param tolerance: image matching tolerance between 0 and 1

        Example:

        .. code-block:: robotframework

            Mouse Click  method=coordinates  100   100
            Mouse Click  CalculatorResults
            Mouse Click  method=image  image=myimage.png  off_x=10  off_y=10  ctype=right
            Mouse Click  method=image  image=myimage.png  tolerance=0.8
            ${elements}  ${other}=     Find Element  class:Button
            FOR  ${element}  IN  @{elements}
                Run Keyword If   ${element}[visible]   Mouse Click  ${element}
            END
            Mouse Click  id:TrickyCheckbox  focus=topleft

        """  # noqa: E501
        self.logger.info("Mouse click: %s", locator)

        if method == "locator":
            target_element = None
            if isinstance(locator, dict):
                target_element = locator
            else:
                element, _ = self.find_element(locator)
                if element and len(element) == 1:
                    target_element = element[0]
                else:
                    raise ValueError(f"Could not find unique element for '{locator}'")
            if target_element is None:
                raise ValueError("Could not find unique element to click")
            if focus == "topleft":
                x, y, _, _ = self._get_element_coordinates(target_element)
                x += 2
                y += 2
            else:
                x, y = self.get_element_center(target_element)
            self.click_type(x + off_x, y + off_y, ctype)
        elif method == "coordinates":
            self.mouse_click_coords(x, y, ctype)
        elif method == "image":
            self.mouse_click_image(image, off_x, off_y, ctype, tolerance)

    def mouse_click_image(
        self,
        template: str,
        off_x: int = 0,
        off_y: int = 0,
        ctype: str = "click",
        tolerance: Optional[float] = None,
    ) -> None:
        """Click at template image on desktop

        :param image: image to click on desktop
        :param off_x: horizontal offset from top left corner to click on
        :param off_y: vertical offset from top left corner to click on
        :param ctype: type of mouse click
        :param tolerance: matching tolerance between 0 and 1

        Example:

        .. code-block:: robotframework

            Mouse Click  image=myimage.png  off_x=10  off_y=10  ctype=right
            Mouse Click  image=myimage.png  tolerance=0.8

        """
        confidence = tolerance * 100.0 if tolerance is not None else None
        locator = ImageLocator(template, confidence=confidence)

        match = Desktop().find_element(locator)

        target = match.center.move(off_x, off_y)
        self.click_type(target.x, target.y, ctype)

    def mouse_click_coords(
        self, x: int, y: int, ctype: str = "click", delay_time: float = None
    ) -> None:
        """Click at coordinates on desktop

        :param x: horizontal coordinate on the windows to click
        :param y: vertical coordinate on the windows to click
        :param ctype: click type "click", "right" or "double", defaults to "click"
        :param delay: delay in seconds after, default is no delay

        Example:

        .. code-block:: robotframework

            Mouse Click Coords  x=450  y=100
            Mouse Click Coords  x=300  y=300  ctype=right
            Mouse Click Coords  x=450  y=100  delay=5.0

        """
        self.click_type(x, y, ctype)
        if delay_time:
            delay(delay_time)

    def get_element(
        self, locator: str, screenshot: bool = False, open_dialog: bool = True
    ) -> Any:
        """Get element by locator.

        :param locator: name of the locator
        :param screenshot: takes element screenshot if True, defaults to False
        :param open_dialog: True if dialog should be reopened, default to True
        :return: element if element was identified, else False

        Example:

        .. code-block:: robotframework

            ${element}  Get Element  CalculatorResults
            ${element}  Get Element  Result      screenshot=True

        """
        self.logger.info("Get element: %s", locator)
        if open_dialog:
            self.open_dialog(self.windowtitle)
        self.dlg.wait("exists enabled visible ready")

        matching_elements, _ = self.find_element(locator)

        if len(matching_elements) == 0:
            self.logger.info(
                "Locator '%s' not found in '%s'.\n",
                locator,
                self.windowtitle,
            )
        elif len(matching_elements) == 1:
            element = matching_elements[0]
            if screenshot:
                self.screenshot(f"locator_{locator}", element=element)
            for key in element.keys():
                self.logger.debug("%s=%s", key, element[key])
            return element
        else:
            # TODO: return more valuable information about what should
            # be matching element ?
            self.logger.info(
                "Locator '%s' matched multiple elements in '%s'. ",
                locator,
                self.windowtitle,
            )
        return False

    def get_element_rich_text(self, locator: str) -> Any:
        """Get value of element `rich text` attribute.

        :param locator: element locator
        :return: `rich_text` value if found, else False

        Example:

        .. code-block:: robotframework

            ${text}  Get Element Rich Text  CalculatorResults

        """
        element = self.get_element(locator)
        if element is not False and "rich_text" in element:
            return element["rich_text"]
        elif element is False:
            self.logger.info("Did not find element with locator: %s", locator)
            return False
        else:
            self.logger.info(
                "Element for locator %s does not have 'rich_text' attribute", locator
            )
            return False

    def get_element_rectangle(self, locator: str, as_dict: bool = False) -> Any:
        # pylint: disable=C0301
        """Get value of element `rectangle` attribute.

        :param locator: element locator
        :param as_dict: return values in a dictionary, default `False`
        :return: (left, top, right, bottom) values if found, else False

        Example:

        .. code-block:: robotframework

            ${left}  ${top}  ${right}  ${bottom}=  Get Element Rectangle  CalculatorResults
            &{coords}  Get Element Rectangle  CalculatorResults  as_dict=True
            Log  top=${coords.top} left=${coords.left}

        """  # noqa: E501
        rectangle = self._get_element_attribute(locator, "rectangle")
        left, top, right, bottom = self._get_element_coordinates(rectangle)
        if as_dict:
            return {"left": left, "top": top, "right": right, "bottom": bottom}
        return left, top, right, bottom

    def _get_element_attribute(self, locator: str, attribute: str) -> Any:
        element = self.get_element(locator)
        if element is not False and attribute in element:
            return element[attribute]
        elif element is False:
            self.logger.info("Did not find element with locator %s", locator)
            return False
        else:
            self.logger.info(
                "Element for locator %s does not have 'visible' attribute", locator
            )
            return False

    def is_element_visible(self, locator: str) -> bool:
        """Is element visible.

        :param locator: element locator
        :return: True if visible, else False

        Example:

        .. code-block:: robotframework

            ${res}=   Is Element Visible  CalculatorResults

        """
        visible = self._get_element_attribute(locator, "visible")
        return bool(visible)

    def is_element_enabled(self, locator: str) -> bool:
        """Is element enabled.

        :param locator: element locator
        :return: True if enabled, else False

        Example:

        .. code-block:: robotframework

            ${res}=   Is Element Enabled  CalculatorResults

        """
        enabled = self._get_element_attribute(locator, "enabled")
        return bool(enabled)

    def menu_select(self, menuitem: str) -> None:
        """Select item from menu

        :param menuitem: name of the menu item

        Example:

        .. code-block:: robotframework

            Open Using Run Dialog   notepad     Untitled - Notepad
            Menu Select             File->Print

        """
        self.logger.info("Menu select: %s", menuitem)
        if self.dlg is None:
            raise ValueError("No dialog open")
        try:
            self.dlg.menu_select(menuitem)
        except AttributeError as e:
            raise MenuItemNotFoundError(
                "Unable to access menu item '%s'" % menuitem
            ) from e

    def wait_for_element(
        self,
        locator: str,
        use_refreshing: bool = False,
        search_criteria: str = None,
        timeout: float = 30.0,
        interval: float = 2.0,
    ) -> Any:
        """Wait for element to appear into the window.

        Can return 1 or more elements matching locator, or raises
        `ElementNotFoundError` if element is not found within timeout.

        :param locator: name of the locator
        :param use_refreshing: wait for element(s) which are not there yet e.g. listbox
         item or popups, default False
        :param search_criteria: criteria by which element is matched
        :param timeout: defines how long to wait for element to appear,
         defaults to 30.0 seconds
        :param interval: how often to poll for element,
         defaults to 2.0 seconds (minimum is 0.5 seconds)

        Example:

        .. code-block:: robotframework

            @{elements}  Wait For Element  CalculatorResults
            @{elements}  Wait For Element  Results   timeout=10  interval=1.5

        """
        self.refresh_window()
        end_time = time.time() + float(timeout)
        interval = max([0.5, interval])
        elements = None
        while time.time() < end_time:
            elements, _ = self.find_element(locator, search_criteria)
            if use_refreshing:
                self.refresh_window()
            if len(elements) > 0:
                break
            if interval >= timeout:
                self.logger.info(
                    "Wait For Element: interval has been set longer than timeout - "
                    "executing one cycle."
                )
                break
            if time.time() >= end_time:
                break
            time.sleep(interval)
        if elements:
            return elements
        raise ElementNotFoundError

    def find_element(self, locator: str, search_criteria: str = None) -> Any:
        """Find element from window by locator and criteria.

        :param locator: name of the locator
        :param search_criteria: criteria by which element is matched
        :return: list of matching elements and locators that were found on the window

        Example:

        .. code-block:: robotframework

            @{elements}   Find Element   CalculatorResults
            Log Many  ${elements[0]}     # list of matching elements
            Log Many  ${elements[1]}     # list of all available locators

        """
        match_type, search_locators = self._parse_locator(locator, search_criteria)
        if self.elements is None:
            controls, elements = self.get_window_elements()
        else:
            controls = self.controls
            elements = self.elements

        matching_elements, locators = [], []

        for ctrl, element in zip(controls, elements):
            match_results = {}
            for criteria, search_locator in search_locators:
                match_results[criteria] = self.is_element_matching(
                    element, search_locator, criteria
                )
            if (match_type == "all" and all(match_results.values())) or (
                match_type == "any" and any(match_results.values())
            ):
                element["control"] = ctrl
                matching_elements.append(element)
        return matching_elements, locators

    def _determine_search_criteria(self, locator: str) -> Any:
        """Check search criteria from locator.

        Possible search criterias:
            - name
            - class / class_name
            - type / control_type
            - id / automation_id
            - partial name (wildcard search for 'name' attribute)
            - any (if none was defined)

        :param locator: name of the locator
        :return: criteria and locator
        """
        if locator.startswith("name:"):
            search_criteria = "name"
            _, locator = locator.split(":", 1)
        elif locator.startswith(("class_name:", "class:")):
            search_criteria = "class_name"
            _, locator = locator.split(":", 1)
        elif locator.startswith(("control_type:", "type:")):
            search_criteria = "control_type"
            _, locator = locator.split(":", 1)
        elif locator.startswith(("automation_id:", "id:")):
            search_criteria = "automation_id"
            _, locator = locator.split(":", 1)
        elif locator.startswith("partial name:"):
            search_criteria = "partial name"
            _, locator = locator.split(":", 1)
        elif locator.startswith("regexp:"):
            search_criteria = "regexp"
            _, locator = locator.split(":", 1)
        else:
            search_criteria = "any"

        return search_criteria, locator

    # TODO: supporting multiple search criterias at same time to identify ONE element
    def _is_element_matching(
        self, itemdict: dict, locator: str, criteria: str, wildcard: bool = False
    ) -> bool:
        if criteria == "regexp":
            name_search = re.search(locator, itemdict["name"])
            class_search = re.search(locator, itemdict["class_name"])
            type_search = re.search(locator, itemdict["control_type"])
            id_search = re.search(locator, itemdict["automation_id"])
            return name_search or class_search or type_search or id_search
        elif criteria != "any" and criteria in itemdict:
            if (wildcard and locator in itemdict[criteria]) or (
                locator == itemdict[criteria]
            ):
                return True
        elif criteria == "any":
            name_search = self.is_element_matching(itemdict, locator, "name")
            class_search = self.is_element_matching(itemdict, locator, "class_name")
            type_search = self.is_element_matching(itemdict, locator, "control_type")
            id_search = self.is_element_matching(itemdict, locator, "automation_id")
            if name_search or class_search or type_search or id_search:
                return True
        elif criteria == "partial name":
            return self.is_element_matching(itemdict, locator, "name", True)
        return False

    # TODO: supporting multiple search criterias at same time to identify ONE element
    def is_element_matching(
        self, itemdict: dict, locator: str, criteria: str, wildcard: bool = False
    ) -> bool:
        """Is element matching. Check if locator is found in `any` field
        or `criteria` field in the window items.

        :param itemDict: dictionary of element items
        :param locator: name of the locator
        :param criteria: criteria on which to match element
        :param wildcard: whether to do reg exp match or not, default False
        :return: True if element is matching locator and criteria, False if not
        """
        return self._is_element_matching(itemdict, locator, criteria, wildcard)

    def get_dialog_rectangle(self, ctrl: Any = None, as_dict: bool = False) -> Any:
        """Get dialog rectangle coordinates

        If `ctrl` is None then get coordinates from `dialog`

        :param ctrl: name of the window control object, defaults to None
        :return: coordinates: left, top, right, bottom

        Example:

        .. code-block:: robotframework

            ${left}  ${top}  ${right}  ${bottom}=  Get Dialog Rectangle
            &{coords}  Get Dialog Rectangle  as_dict=True
            Log  top=${coords.top} left=${coords.left}

        """
        if ctrl:
            rect = ctrl.element_info.rectangle
        elif self.dlg:
            rect = self.dlg.element_info.rectangle
        else:
            raise ValueError("No dialog open")

        if as_dict:
            return {
                "left": rect.left,
                "top": rect.top,
                "right": rect.right,
                "bottom": rect.bottom,
            }
        else:
            return rect.left, rect.top, rect.right, rect.bottom

    def get_element_center(self, element: dict) -> Any:
        """Get element center coordinates

        :param element: dictionary of element items
        :return: coordinates, x and y

        Example:

        .. code-block:: robotframework

            @{element}   Find Element  CalculatorResults
            ${x}  ${y}=  Get Element Center  ${elements[0][0]}

        """
        return self.calculate_rectangle_center(element["rectangle"])

    def click_type(
        self, x: int = None, y: int = None, click_type: str = "click"
    ) -> None:
        """Mouse click on coordinates x and y.

        Default click type is `click` meaning `left`

        :param x: horizontal coordinate for click, defaults to None
        :param y: vertical coordinate for click, defaults to None
        :param click_type: "click", "right" or "double", defaults to "click"
        :raises ValueError: if coordinates are not valid

        Example:

        .. code-block:: robotframework

            Click Type  x=450  y=100
            Click Type  x=450  y=100  click_type=right
            Click Type  x=450  y=100  click_type=double

        """
        self.logger.info("Click type '%s' at (%s, %s)", click_type, x, y)
        if (x is None and y is None) or (x < 0 or y < 0):
            raise ValueError(f"Can't click on given coordinates: ({x}, {y})")
        if click_type == "click":
            pywinauto.mouse.click(coords=(x, y))
        elif click_type == "double":
            pywinauto.mouse.double_click(coords=(x, y))
        elif click_type == "right":
            pywinauto.mouse.right_click(coords=(x, y))

    def get_window_elements(
        self,
        screenshot: bool = False,
        element_json: bool = False,
        outline: bool = False,
    ) -> Any:
        # pylint: disable=C0301
        """Get element information about all window dialog controls
        and their descendants.

        :param screenshot: save element screenshot if True, defaults to False
        :param element_json: save element json if True, defaults to False
        :param outline: highlight elements if True, defaults to False
        :return: all controls and all elements

        Example:

        .. code-block:: robotframework

            @{elements}   Get Window Elements
            Log Many      ${elements[0]}     # list of all available locators
            Log Many      ${elements[1]}     # list of matching elements
            @{elements}   Get Window Elements  screenshot=True  element_json=True  outline=True

        """  # noqa: E501
        ctrls = self._get_all_window_controls()
        elements, controls = [], []
        for _, ctrl in enumerate(ctrls):
            try:
                if not hasattr(ctrl, "element_info"):
                    continue
            except COMError as ce:
                self.logger.info("Got COM error: %s", str(ce))
                continue

            filename = clean_filename(
                f"locator_{self.windowtitle}_ctrl_{ctrl.element_info.name}"
            )

            if screenshot and len(ctrl.element_info.name) > 0:
                self.screenshot(filename, ctrl=ctrl, overwrite=True)
            if outline:
                ctrl.draw_outline(colour="red", thickness=4)
                delay(0.2)
                ctrl.draw_outline(colour=0x000000, thickness=4)

            element = self._parse_element_attributes(element=ctrl)
            if element_json:
                write_element_info_as_json(element, filename)

            controls.append(ctrl)
            elements.append(element)

        if element_json:
            write_element_info_as_json(
                elements, clean_filename(f"locator_{self.windowtitle}_all_elements")
            )

        return controls, elements

    def _get_all_window_controls(self):
        if self.dlg is None:
            raise ValueError("No dialog open")

        ctrls = [self.dlg]
        try:
            if hasattr(self.dlg, "descendants"):
                ctrls += self.dlg.descendants()
        except COMError as ce:
            self.logger.info("Got COM error: %s", str(ce))
        return ctrls

    def _get_element_coordinates(self, element: Any) -> Any:
        """Get element coordinates from pywinauto object.

        :param rectangle: item containing rectangle information
        :return: coordinates: left, top, right, bottom
        """
        left = 0
        top = 0
        right = 0
        bottom = 0
        if isinstance(element, pywinauto.win32structures.RECT):
            left = element.left
            top = element.top
            right = element.right
            bottom = element.bottom
        elif isinstance(element, dict) and "rectangle" not in element.keys():
            left = element.left
            top = element.top
            right = element.right
            bottom = element.bottom
        else:
            if isinstance(element, dict) and "rectangle" in element.keys():
                rectangle = element["rectangle"]
            else:
                rectangle = element
            left, top, right, bottom = map(
                int,
                re.match(
                    r"\(L([-]?\d+).*T([-]?\d+).*R([-]?\d+).*B([-]?\d+)\)",
                    str(rectangle),
                ).groups(),
            )
        return left, top, right, bottom

    def screenshot(
        self,
        filename: str,
        element: dict = None,
        ctrl: Any = None,
        desktop: bool = False,
        overwrite: bool = True,
    ) -> None:
        """Save screenshot into filename.

        :param filename: name of the file
        :param element: take element screenshot, defaults to None
        :param ctrl: take control screenshot, defaults to None
        :param desktop: take desktop screenshot if True, defaults to False
        :param overwrite: overwrite existing image (deprecated, always True)

        Example:

        .. code-block:: robotframework

            @{element}   Find Element  CalculatorResults
            Screenshot   element.png   ${elements[0][0]}
            Screenshot   desktop.png   desktop=True
            Screenshot   desktop.png   desktop=True  overwrite=True

        """
        del overwrite  # Always overwrite

        if desktop:
            region = None
        elif element:
            region = self._get_element_coordinates(element["rectangle"])
        elif ctrl:
            region = self.get_dialog_rectangle(ctrl)
        else:
            region = self.get_dialog_rectangle()

        if region is not None:
            region = Region(*region)

        Desktop().take_screenshot(path=filename, locator=region)

    def _parse_element_attributes(self, element: dict) -> dict:
        """Return filtered element dictionary for an element.

        :param element: should contain `element_info` attribute
        :return: dictionary containing element attributes
        """
        if element is None and "element_info" not in element:
            self.logger.warning(
                "%s is none or does not have element_info attribute", element
            )
            return None

        element_dict = self._prepare_element_dict(element)
        element_info = element.element_info
        element_attributes = [a for a in dir(element_info) if not a.startswith("_")]

        for attr in element_attributes:
            try:
                attr_value = getattr(element_info, attr)
                if attr == "parent":
                    element_dict["parent"] = getattr(attr_value, "control_type", None)
                else:
                    element_dict[attr] = (
                        attr_value() if callable(attr_value) else str(attr_value)
                    )
            except TypeError:
                pass
            except NotImplementedError:
                pass
            except COMError as ce:
                self.logger.info("Got COM error: %s", str(ce))

        return self._clean_element_dict(element_dict)

    def _prepare_element_dict(self, element):
        element_dict = {}

        element_dict["object"] = element
        try:
            element_dict["legacy"] = (
                element.legacy_properties()
                if hasattr(element, "legacy_properties")
                else None
            )
        except AttributeError:
            pass
        return element_dict

    def _clean_element_dict(self, element_dict):
        attributes_to_remove = [
            # "automation_id",
            "children",
            # "class_name",
            # "control_id",
            # "control_type",
            "descendants",
            "dump_window",
            "element",
            # "enabled",
            "filter_with_depth",
            "framework_id",
            "from_point",
            # "handle",
            "has_depth",
            "iter_children",
            "iter_descendants",
            # "name",
            # "parent",
            # "process_id",
            # "rectangle",
            # "rich_text",
            # "runtime_id",
            "set_cache_strategy",
            "top_from_point",
            # "visible",
        ]

        for attr in attributes_to_remove:
            element_dict.pop(attr, None)

        return element_dict

    def put_system_to_sleep(self) -> None:
        """Put Windows into sleep mode

        Example:

        .. code-block:: robotframework

            Put System To Sleep

        """
        access = win32security.TOKEN_ADJUST_PRIVILEGES | win32security.TOKEN_QUERY
        htoken = win32security.OpenProcessToken(win32api.GetCurrentProcess(), access)
        if htoken:
            priv_id = win32security.LookupPrivilegeValue(
                None, win32security.SE_SHUTDOWN_NAME
            )
            win32security.AdjustTokenPrivileges(
                htoken, 0, [(priv_id, win32security.SE_PRIVILEGE_ENABLED)]
            )
            ctypes.windll.powrprof.SetSuspendState(False, True, True)
            win32api.CloseHandle(htoken)

    def lock_screen(self) -> None:
        """Put windows into lock mode

        Example:

        .. code-block:: robotframework

            Lock Screen
        """
        ctypes.windll.User32.LockWorkStation()

    def log_in(self, username: str, password: str, domain: str = ".") -> str:
        """Log into Windows `domain` with `username` and `password`.

        :param username: name of the user
        :param password: password of the user
        :param domain: windows domain for the user, defaults to "."
        :return: handle

        Example:

        .. code-block:: robotframework

            Log In  username=myname  password=mypassword  domain=company
        """
        return win32security.LogonUser(
            username,
            domain,
            password,
            win32con.LOGON32_LOGON_INTERACTIVE,
            win32con.LOGON32_PROVIDER_DEFAULT,
        )

    def _validate_target(self, target: dict, target_locator: str) -> Any:
        target_x = target_y = 0
        if target_locator is not None:
            self.switch_to_application(target["id"])
            target_elements, _ = self.find_element(target_locator)
            if len(target_elements) == 0:
                raise ValueError(
                    ("Target element was not found by locator '%s'" % target_locator)
                )
            elif len(target_elements) > 1:
                raise ValueError(
                    (
                        "Target element matched more than 1 element (%d) "
                        "by locator '%s'" % (len(target_elements), target_locator)
                    )
                )
            target_x, target_y = self.calculate_rectangle_center(
                target_elements[0]["rectangle"]
            )
        else:
            target_x, target_y = self.calculate_rectangle_center(
                target["dlg"].rectangle()
            )
        return target_x, target_y

    def _select_elements_for_drag(
        self, src: dict, src_locator: str, origin="middle"
    ) -> Any:
        self.switch_to_application(src["id"])
        source_elements, _ = self.find_element(src_locator)
        if len(source_elements) == 0:
            raise ValueError(
                ("Source elements where not found by locator '%s'", src_locator)
            )
        selections = []
        source_min_left = 99999
        source_max_right = -1
        source_min_top = 99999
        source_max_bottom = -1
        for elem in source_elements:
            left, top, right, bottom = self._get_element_coordinates(elem["rectangle"])
            if left < source_min_left:
                source_min_left = left
            if right > source_max_right:
                source_max_right = right
            if top < source_min_top:
                source_min_top = top
            if bottom > source_max_bottom:
                source_max_bottom = bottom
            mid_x = int((right - left) / 2) + left
            mid_y = int((bottom - top) / 2) + top
            if origin == "middle":
                selections.append((mid_x, mid_y))
            elif origin == "topleft":
                selections.append((left, top))
        source_x = int((source_max_right - source_min_left) / 2) + source_min_left
        source_y = int((source_max_bottom - source_min_top) / 2) + source_min_top
        return selections, source_x, source_y

    def drag_and_drop(
        self,
        src: Any,
        target: Any,
        src_locator: str,
        target_locator: str = None,
        handle_ctrl_key: bool = False,
        drop_delay: float = 2.0,
        origin: str = "middle",
    ) -> None:
        # pylint: disable=C0301
        """Drag elements from source and drop them on target.

        Please note that if CTRL is not pressed down during drag and drop then
        operation is MOVE operation, on CTRL down the operation is COPY operation.

        There will be also overwrite notification if dropping over existing files.

        :param src: application object or instance id
        :param target: application object or instance id
        :param src_locator: elements to move
        :param handle_ctrl_key: True if keyword should press CTRL down dragging
        :param drop_delay: how many seconds to wait until releasing mouse drop,
         default 2.0
        :raises ValueError: on validation errors

        Example:

        .. code-block:: robotframework

            ${app1}=        Open Using Run Dialog    explorer.exe{VK_SPACE}C:\\workfiles\\movethese   movethese
            ${app2}=        Open Using Run Dialog    wordpad.exe   Document - WordPad
            Drag And Drop   ${app1}   ${app2}   regexp:testfile_\\d.txt  name:Rich Text Window   handle_ctrl_key=${True}
            Drag And Drop   ${app1}   ${app1}   regexp:testfile_\\d.txt  name:subdir  handle_ctrl_key=${True}

        """  # noqa : E501
        if isinstance(src, int):
            src = self.get_app(src)
        if isinstance(target, int):
            target = self.get_app(target)

        single_application = src["app"] == target["app"]
        selections, source_x, source_y = self._select_elements_for_drag(
            src, src_locator, origin
        )
        target_x, target_y = self._validate_target(target, target_locator)

        self.logger.info(
            "Dragging %d elements from (%d,%d) to (%d,%d)",
            len(selections),
            source_x,
            source_y,
            target_x,
            target_y,
        )

        try:
            if handle_ctrl_key:
                self.send_keys("{VK_LCONTROL down}")
                delay(0.2)

            # Select elements by mouse clicking
            if not single_application:
                self.restore_dialog(src["windowtitle"])
            for idx, selection in enumerate(selections):
                self.logger.debug("Selecting item %d by mouse_click", idx)
                # pywinauto.mouse.click(coords=(selection[0]+5, selection[1]+5))
                self.mouse_click_coords(selection[0] + 5, selection[1] + 5)

            # Start drag from the last item
            pywinauto.mouse.press(coords=(source_x, source_y))
            delay(0.5)
            if not single_application:
                self.restore_dialog(target["windowtitle"])
            pywinauto.mouse.move(coords=(target_x, target_y))

            self.logger.debug("Cursor position: %s", win32api.GetCursorPos())
            delay(drop_delay)
            self.mouse_click_coords(target_x, target_y)
            pywinauto.mouse.click(coords=(target_x, target_y))

            # if action_required:
            self.send_keys("{ENTER}")
            if handle_ctrl_key:
                self.send_keys("{VK_LCONTROL up}")
                delay(0.5)
            # Deselect elements by mouse clicking
            for selection in selections:
                self.logger.debug("Deselecting item by mouse_click")
                self.mouse_click_coords(selection[0] + 5, selection[1] + 5)
        finally:
            self.send_keys("{VK_LCONTROL up}")

    def calculate_rectangle_center(self, rectangle: Any) -> Any:
        """Calculate x and y center coordinates from rectangle.

        :param rectangle: element rectangle coordinates
        :return: x and y coordinates of rectangle center

        Example:

        .. code-block:: robotframework

            Open Using Run Dialog   calc  Calculator
            &{rect}=        Get Element Rectangle    CalculatorResults
            ${x}  ${y}=     Calculate Rectangle Center   ${rect}
        """
        left, top, right, bottom = self._get_element_coordinates(rectangle)
        x = int((right - left) / 2) + left
        y = int((bottom - top) / 2) + top
        return x, y

    def get_window_list(self):
        """Get list of open windows

        Window dictionaries contain:

        - automation_id
        - control_id
        - title
        - pid
        - handle
        - is_active
        - keyboard_focus
        - rectangle

        :return: list of window dictionaries

        Example:

        .. code-block:: robotframework

            @{windows}    Get Window List
            FOR  ${window}  IN  @{windows}
                Log Many  ${window}
            END
        """
        windows = pywinauto.Desktop(backend=self._backend).windows()
        window_list = []
        for w in windows:
            try:
                left, top, right, bottom = self._get_element_coordinates(w.rectangle())
                window_list.append(
                    {
                        "automation_id": w.automation_id(),
                        "control_id": w.control_id(),
                        "title": w.window_text(),
                        "pid": w.process_id(),
                        "handle": w.handle,
                        "is_active": w.is_active(),
                        "keyboard_focus": w.has_keyboard_focus(),
                        "rectangle": [left, top, right, bottom],
                        "object": w,
                    }
                )
            except Exception as e:  # pylint: disable=broad-except
                self.logger.debug(str(e))
        return window_list

    def refresh_window(self):
        """Get controls and elements for current windows.

        Should be called always when window content changes on
        Windows desktop.

        :return: controls (list) and elements (list)
        """
        self.logger.debug("Refresh window")
        controls, elements = self.get_window_elements()
        self.elements = elements
        self.controls = controls
        return controls, elements

    def _parse_locator(self, locator: str, search_criteria: str):
        regex = rf"({':|'.join(WINDOWS_LOCATOR_STRATEGIES.keys())}:|or|and)('{{1}}(.+)'{{1}}|(\S+))?"  # noqa: E501
        parts = re.finditer(regex, locator, re.IGNORECASE)

        locators = []
        match_type = "all"

        for part in parts:
            groups = part.groups()
            if groups[0].lower() == "or":
                match_type = "any"
            elif groups[0].lower() == "and":
                pass
            else:
                strategy, _ = groups[0].split(":")
                value = groups[2] if groups[2] else groups[3]
                locators.append([WINDOWS_LOCATOR_STRATEGIES[strategy], value])

        # Add some strategies if there aren't other valid strategies
        if not locators:
            match_type = "any"
            locators.append(["name", locator])
            locators.append(["automation_id", locator])
            locators.append([search_criteria, locator])
        return match_type, locators

    def set_automation_speed(self, speed: Speed = Speed.DEFAULT):
        """Set global automation timings

        :param speed: possible values 'default', 'fast' or 'slow'
        """
        speed = to_speed(speed)

        if speed == Speed.DEFAULT:
            pywinauto.timings.Timings.defaults()
        elif speed == Speed.SLOW:
            pywinauto.timings.Timings.slow()
        elif speed == Speed.FAST:
            pywinauto.timings.Timings.fast()



# pylint: disable=too-many-function-args
import json
import logging
import os
import random
import time
import urllib.parse as urlparse
from json import JSONDecodeError  # pylint: disable=no-name-in-module
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Union

import requests
from requests.exceptions import HTTPError
from robot.libraries.BuiltIn import BuiltIn, RobotNotRunningError
from tenacity import (
    before_sleep_log,
    retry,
    retry_if_exception,
    stop_after_attempt,
    wait_random_exponential,
)

from RPA.JSON import JSONType
from RPA.RobotLogListener import RobotLogListener


PathType = Union[Path, str]

DEBUG_ON = bool(os.getenv("RPA_DEBUG_API"))
log_to_console = BuiltIn().log_to_console


def url_join(*parts):
    """Join parts of URL and handle missing/duplicate slashes."""
    return "/".join(str(part).strip("/") for part in parts)


def json_dumps(payload: JSONType, **kwargs):
    """Create JSON string in UTF-8 encoding."""
    kwargs.setdefault("ensure_ascii", False)
    return json.dumps(payload, **kwargs)


def is_json_equal(left: JSONType, right: JSONType):
    """Deep-compare two output JSONs."""
    return json_dumps(left, sort_keys=True) == json_dumps(right, sort_keys=True)


def truncate(text: str, size: int):
    """Truncate a string from the middle."""
    if len(text) <= size:
        return text

    ellipsis = " ... "
    segment = (size - len(ellipsis)) // 2
    return text[:segment] + ellipsis + text[-segment:]


def resolve_path(path: str) -> Path:
    """Resolve a string-based path, and replace variables."""
    try:
        safe = str(path).replace("\\", "\\\\")
        path = BuiltIn().replace_variables(safe)
    except RobotNotRunningError:
        pass

    return Path(path).expanduser().resolve()


def log_more(message, *args, func=logging.debug):
    func(message, *args)
    if DEBUG_ON:
        log_to_console(str(message) % args)


def get_dot_value(source: Dict, key: str) -> Any:
    """Returns the end value from `source` dictionary given `key` traversal."""
    keys = key.split(".")
    value = source
    for _key in keys:
        value = value.get(_key)
    return value


def set_dot_value(source: Dict, key: str, *, value: Any):
    """Sets the end `value` into `source` dictionary given `key` destination."""
    keys = key.rsplit(".", 1)  # one or at most two parts
    if len(keys) == 2:
        source = get_dot_value(source, keys[0])
    source[keys[-1]] = value


class RequestsHTTPError(HTTPError):
    """Custom `requests` HTTP error with status code and message."""

    def __init__(
        self, *args, status_code: int = 0, status_message: str = "Error", **kwargs
    ):
        super().__init__(*args, **kwargs)

        self.status_code = status_code
        self.status_message = status_message


class Requests:
    """Wrapper over `requests` 3rd-party with error handling and retrying support."""

    def __init__(self, route_prefix: str, default_headers: Optional[dict] = None):
        self._route_prefix = route_prefix
        self._default_headers = default_headers

    def handle_error(self, response: requests.Response):
        resp_status_code = response.status_code
        log_func = logging.critical if resp_status_code // 100 == 5 else logging.debug
        log_more(
            "API response: %s %r", resp_status_code, response.reason, func=log_func
        )
        if response.ok:
            return

        fields = {}
        try:
            fields = response.json()
            while not isinstance(fields, dict):
                # For some reason we might still get a string from the deserialized
                # retrieved JSON payload. If a dictionary couldn't be obtained at all,
                # it will end up raising `RequestsHTTPError`.
                fields = json.loads(fields)
        except (JSONDecodeError, ValueError, TypeError):
            # No `fields` dictionary can be obtained at all.
            log_more("No fields were returned by the server", func=logging.critical)
            try:
                response.raise_for_status()
            except Exception as exc:  # pylint: disable=broad-except
                log_more(exc, func=logging.exception)
                raise RequestsHTTPError(exc, status_code=resp_status_code) from exc

        err_status_code = 0
        status_message = "Error"
        try:
            err_status_code = int(fields.get("status", resp_status_code))
            status_message = fields.get("error", {}).get("code", "Error")
            reason = fields.get("message") or fields.get("error", {}).get(
                "message", response.reason
            )
            raise HTTPError(f"{err_status_code} {status_message}: {reason}")
        except Exception as exc:  # pylint: disable=broad-except
            log_more(exc, func=logging.exception)
            raise RequestsHTTPError(
                str(fields), status_code=err_status_code, status_message=status_message
            ) from exc

    # pylint: disable=no-self-argument
    def _needs_retry(exc: BaseException) -> bool:
        # Don't retry on some specific error codes or messages.

        # https://www.restapitutorial.com/httpstatuscodes.html
        # 400 - payload is bad and needs to be changed
        # 401 - missing auth bearer token
        # 403 - auth is in place, but not allowed (insufficient privileges)
        # 409 - payload not good for the affected resource
        no_retry_codes = [400, 401, 403, 409]
        no_retry_messages = []

        if isinstance(exc, RequestsHTTPError):
            exc: RequestsHTTPError
            if (
                exc.status_code in no_retry_codes
                or exc.status_message in no_retry_messages
            ):
                return False

            if exc.status_code == 429:
                # We hit the rate limiter, so sleep extra.
                seconds = random.uniform(1, 3)
                log_more("Rate limit hit, sleeping: %fs", seconds, func=logging.warning)
                time.sleep(seconds)

        return True

    # pylint: disable=no-self-argument,no-method-argument
    def _before_sleep_log():
        logger = logging.root
        logger_log = logger.log

        def extensive_log(level, msg, *args, **kwargs):
            logger_log(level, msg, *args, **kwargs)
            if DEBUG_ON:
                log_to_console(str(msg) % args)

        # Monkeypatch inner logging function so it produces an exhaustive log when
        # used under the before-sleep logging utility in `tenacity`.
        logger.log = extensive_log
        return before_sleep_log(logger, logging.DEBUG, exc_info=True)

    @retry(
        # Retry until either succeed or trying for the fifth time and still failing.
        # So sleep and retry for 4 times at most.
        stop=stop_after_attempt(5),
        # If the exception is no worth retrying or the number of tries is depleted,
        # then re-raise the last raised exception.
        reraise=True,
        # Decide if the raised exception needs retrying or not.
        retry=retry_if_exception(_needs_retry),
        # Produce debugging logging prior to each time we sleep & re-try.
        before_sleep=_before_sleep_log(),
        # Sleep between the tries with a random float amount of seconds like so:
        # 1. [0, 2]
        # 2. [0, 4]
        # 3. [0, 5]
        # 4. [0, 5]
        wait=wait_random_exponential(multiplier=2, max=5),
    )
    def _request(
        self,
        verb: Callable[..., requests.Response],
        url: str,
        *args,
        _handle_error: Optional[Callable[[requests.Response], None]] = None,
        _sensitive: bool = False,
        headers: Optional[dict] = None,
        **kwargs,
    ) -> requests.Response:
        # Absolute URLs override the prefix, so they are safe to be sent as they'll be
        # the same after joining.
        url = urlparse.urljoin(self._route_prefix, url)
        headers = headers if headers is not None else self._default_headers
        handle_error = _handle_error or self.handle_error

        url_for_log = url
        if _sensitive:
            # Omit query from the URL since might contain sensitive info.
            split = urlparse.urlsplit(url_for_log)
            url_for_log = urlparse.urlunsplit(
                [split.scheme, split.netloc, split.path, "", split.fragment]
            )
        log_more("%s %r", verb.__name__.upper(), url_for_log)
        response = verb(url, *args, headers=headers, **kwargs)
        handle_error(response)
        return response

    # CREATE
    def post(self, *args, **kwargs) -> requests.Response:
        return self._request(requests.post, *args, **kwargs)

    # RETRIEVE
    def get(self, *args, **kwargs) -> requests.Response:
        return self._request(requests.get, *args, **kwargs)

    # UPDATE
    def put(self, *args, **kwargs) -> requests.Response:
        return self._request(requests.put, *args, **kwargs)

    # DELETE
    def delete(self, *args, **kwargs) -> requests.Response:
        return self._request(requests.delete, *args, **kwargs)


def protect_keywords(base: str, keywords: List[str]):
    """Protects from logging a list of `keywords` relative to `base`."""
    to_protect = [f"{base}.{keyword}" for keyword in keywords]
    listener = RobotLogListener()
    listener.register_protected_keywords(to_protect)


def get_output_dir(default: Optional[PathType] = "output") -> Path:
    """Returns the output directory path."""
    try:
        output_dir = BuiltIn().get_variable_value("${OUTPUT_DIR}", default=default)
    except RobotNotRunningError:
        output_dir = default
    return Path(output_dir).expanduser().resolve()



try:
    from importlib import metadata
except ImportError:
    import importlib_metadata as metadata

# this should import the version from pyproject.toml
# it uses a catch-all try block in case the package is partially loaded.
try:
    __version__ = metadata.version("rpaframework")
except Exception:  # pylint: disable=broad-except
    __version__ = "0.0.0"



import RPA.Smartsheet as ss
from RPA.Tables import Table

# Import mock data
from .smartsheet_vars import (
    ITEM_COLUMN_ID,
    MOCK_ACCESS_TOKEN,
    MOCK_IMAGE_ATTACHMENT,
    MOCK_SHEET_JSON,
    MOCK_TEXT_ATTACHMENT,
    MOCKED_ADD_ROW_SUCCESS,
    MOCKED_ATTACHMENT_LIST,
    MOCKED_CREATE_SHEET_SUCCESS,
    MOCKED_ROW,
    MOCKED_ROW_ATTACHMENTS,
    MOCKED_ROW_NO_ATTACHMENTS,
    MOCKED_SEARCH_RESULTS,
    MOCKED_UPDATE_ROW_SUCCESS,
    NAME_COLUMN_ID,
    NEW_ROWS,
    NEW_SHEET_COLUMNS,
    NEW_SHEET_NAME,
    UPDATED_ROW_1,
    UPDATED_ROWS,
    ZIP_COLUMN_ID,
)
from itertools import chain
from json.encoder import JSONEncoder
from pathlib import Path
from typing import Iterable, List, Union

import pytest
import smartsheet.smartsheet as smart_sdk
from mock import MagicMock
from pytest_mock import MockerFixture
from requests.models import PreparedRequest, Response

OperationResult = smart_sdk.OperationResult
OperationErrorResult = smart_sdk.OperationErrorResult

Smartsheet = ss.Smartsheet

# You can set a personal testing token via the smartsheet_testvars.py file
try:
    from .smartsheet_testvars import ACCESS_TOKEN

    VARS = True
except ImportError:
    VARS = False



# Testing Constants
TEMP_DIR = Path(__file__).parent.parent / "results"


# Need to skip if smartsheet_testvars could not be imported
@pytest.mark.skip("Test only used for debugging")
def test_debugging():
    """This test should always be skipped for automated tests."""
    lib = Smartsheet(access_token=ACCESS_TOKEN)
    orders = lib.get_sheet(sheet_name="orders", native=True)
    _ = lib.get_cell_history(row=orders.rows[0], column=orders.columns[0])
    attachments = lib.list_attachments()
    for a in attachments:
        lib.download_attachment(a, TEMP_DIR)


@pytest.fixture
def library() -> Smartsheet:
    return Smartsheet()


@pytest.fixture
def authorized_lib(
    library: Smartsheet, mocker: MockerFixture, iter: int = 1
) -> Smartsheet:
    library.set_access_token(MOCK_ACCESS_TOKEN.format(iter))
    return library


@pytest.fixture
def lib_with_sheet(authorized_lib: Smartsheet, mocker: MockerFixture) -> Smartsheet:
    _patch_response(authorized_lib, mocker, MOCK_SHEET_JSON)
    authorized_lib.get_sheet(7340596407887748, include="rowId")
    return authorized_lib


def _create_json_response(return_value: dict) -> MagicMock:
    mock_response = MagicMock(Response)
    mock_response.status_code = 200
    mock_response.reason = "OK"
    mock_response.headers = {"Content-Type": "application/json"}
    mock_response.json.return_value = return_value
    text = JSONEncoder().encode(return_value)
    mock_response.text = text
    mock_response.content.decode.return_value = text
    mock_response.request = MagicMock(PreparedRequest)
    mock_response.request.method = "TEST"
    mock_response.request.url = "TEST"
    mock_response.request.body = None
    return mock_response


def _create_file_response(
    expected_content: Union[str, bytes],
) -> MagicMock:
    if isinstance(expected_content, str):
        content = expected_content.encode()
    elif isinstance(expected_content, bytes):
        content = expected_content
    else:
        raise TypeError("Expected content must be str or bytes")
    file_response = MagicMock()
    file_response.content = content
    file_response.status_code = 200
    file_response.iter_content.return_value = iter([content])
    return file_response


def _create_file_json(
    expected_filename: str,
    mime_type: str = "text/plain;charset=UTF-8",
) -> dict:
    return {
        "id": 123,
        "name": str(expected_filename),
        "url": "http://localhost",
        "attachmentType": "FILE",
        "mimeType": str(mime_type),
        "urlExpiresInMillis": 120000,
        "sizeInKb": 1,
        "parentType": "ROW",
        "parentId": 456,
        "createdAt": "2023-02-22T22:46:44Z",
        "createdBy": {"email": "markmonkey@robocorp.com"},
    }


def _patch_file_response(
    library: Smartsheet,
    mocker: MockerFixture,
    expected_filename: str,
    expected_content: Union[str, bytes],
    mime_type: str = "text/plain;charset=UTF-8",
) -> MagicMock:
    _patch_response(library, mocker, _create_file_json(expected_filename, mime_type))
    mock_get = mocker.patch("smartsheet.attachments.requests.get")
    mock_get.return_value = _create_file_response(expected_content)
    return mock_get


def _patch_multiple_file_responses(
    library: Smartsheet,
    mocker: MockerFixture,
    expected_filenames: List[str],
    expected_contents: List[Union[str, bytes]],
    mime_types: List[str],
    send_patch: MagicMock = None,
) -> MagicMock:
    """Creates multiple file responses for a single request.

    Warning: Send patch is not returned, but is modified in place.
    """
    json_responses = []
    for filename, mime_type in zip(expected_filenames, mime_types):
        file_json = _create_file_json(filename, mime_type)
        json_responses.append(_create_json_response(file_json))

    if send_patch is None:
        _patch_multiple_responses(library, mocker, json_responses)
    else:
        _extend_patch(send_patch, json_responses)
    file_responses = []
    for content in expected_contents:
        file_responses.append(_create_file_response(content))
    mock_get = mocker.patch("smartsheet.attachments.requests.get")
    mock_get.side_effect = file_responses
    return mock_get


def _patch_response(
    library: Smartsheet, mocker: MockerFixture, return_value: dict
) -> MagicMock:
    mock_response = _create_json_response(return_value)
    config = {"return_value": mock_response}
    return mocker.patch.object(library.smart._session, "send", **config)


def _patch_multiple_responses(
    library: Smartsheet, mocker: MockerFixture, mocked_responses: Iterable[MagicMock]
) -> MagicMock:
    config = {"side_effect": mocked_responses}
    return mocker.patch.object(library.smart._session, "send", **config)


def _extend_patch(patch: MagicMock, mocked_responses: Iterable[MagicMock]) -> MagicMock:
    """This function will extend the side_effect of a patch with new responses.

    It does this by chaining the current side_effect with the new responses.

    :param patch: The patch to extend.
    :param mocked_responses: The responses to add to the patch.
    """
    current_side_effect = patch.side_effect
    if current_side_effect is None:
        patch.side_effect = mocked_responses
    else:
        patch.side_effect = chain(current_side_effect, mocked_responses)
    return patch


def _wrap_lib(lib_with_sheet: Smartsheet, mocker: MockerFixture) -> MagicMock:
    func_wrap = mocker.Mock(wraps=lib_with_sheet.smart)
    func_patch = mocker.patch.object(lib_with_sheet, "smart", func_wrap)
    return func_patch


def test_authorization(library: Smartsheet, mocker: MockerFixture) -> None:
    mock_client = mocker.patch("RPA.Smartsheet.SmartSDK", autospec=True)

    library.set_access_token(MOCK_ACCESS_TOKEN)

    mock_client.assert_called()


def test_get_application_constants(
    authorized_lib: Smartsheet, mocker: MockerFixture
) -> None:
    expected_response = {"supportedLocales": ["en_US"], "serverVersion": "207.0.0"}
    patch = _patch_response(authorized_lib, mocker, expected_response)

    server_info = authorized_lib.get_application_constants()

    patch.assert_called()
    assert list(server_info.supported_locales) == ["en_US"]


def test_get_me(authorized_lib: Smartsheet, mocker: MockerFixture) -> None:
    expected_response = {
        "id": 123123,
        "email": "markmonkey@robocorp.com",
        "locale": "en_US",
        "timeZone": "US/Pacific",
        "account": {
            "name": "markmonkey@robocorp.com (Developer)",
            "id": 789789,
        },
        "admin": True,
        "licensedSheetCreator": True,
        "groupAdmin": True,
        "resourceViewer": True,
        "alternateEmails": [],
        "sheetCount": 25,
        "lastLogin": "2023-02-28T18:15:56Z",
        "title": "",
        "department": "",
        "company": "",
        "workPhone": "",
        "mobilePhone": "",
        "role": "",
    }
    patch = _patch_response(authorized_lib, mocker, expected_response)

    me = authorized_lib.get_current_user()

    patch.assert_called()
    assert me.email == expected_response["email"]


def test_get_sheet(authorized_lib: Smartsheet, mocker: MockerFixture) -> None:
    patch = _patch_response(authorized_lib, mocker, MOCK_SHEET_JSON)

    sheet = authorized_lib.get_sheet(123, native=True)

    patch.assert_called()
    assert sheet.id == MOCK_SHEET_JSON["id"]
    assert [c.title for c in sheet.columns] == [
        c["title"] for c in MOCK_SHEET_JSON["columns"]
    ]
    for row in sheet.rows:
        assert row.id in [r["id"] for r in MOCK_SHEET_JSON["rows"]]


def test_convert_sheet_to_table(
    lib_with_sheet: Smartsheet, mocker: MockerFixture
) -> None:
    table = lib_with_sheet.convert_sheet_to_table()

    assert isinstance(table, Table)
    expected_columns = MOCK_SHEET_JSON["columns"].copy()
    expected_columns.append({"title": "rowId"})
    assert table.columns == [c["title"] for c in expected_columns]
    for table_row, sheet_row in zip(table, MOCK_SHEET_JSON["rows"]):
        expected_row = [cell["value"] for cell in sheet_row["cells"].copy()]
        expected_row.append(sheet_row["id"])
        assert list(table_row.values()) == expected_row


def test_download_attachment(lib_with_sheet: Smartsheet, mocker: MockerFixture) -> None:
    expected_filename = "dummy.txt"
    expected_content = "dummy file content"
    mock_get = _patch_file_response(
        lib_with_sheet, mocker, expected_filename, expected_content
    )

    downloaded_file = lib_with_sheet.download_attachment(123, TEMP_DIR)

    mock_get.assert_called()
    assert downloaded_file
    assert downloaded_file.exists()
    assert downloaded_file.read_text() == str(expected_content)


def test_get_sheet_with_attachments(
    authorized_lib: Smartsheet, mocker: MockerFixture
) -> None:
    mock_sheet = MOCK_SHEET_JSON.copy()
    mock_sheet["rows"][0]["attachments"] = MOCKED_ATTACHMENT_LIST
    send_responses = []
    send_responses.append(_create_json_response(mock_sheet))
    send_responses.append(_create_json_response(MOCKED_ROW_ATTACHMENTS))
    send_patch = _patch_multiple_responses(authorized_lib, mocker, send_responses)
    attachments = [
        MOCK_IMAGE_ATTACHMENT,
        MOCK_TEXT_ATTACHMENT,
    ]
    attachment_patches = _patch_multiple_file_responses(
        authorized_lib,
        mocker,
        [a["expected_filename"] for a in attachments],
        [a["expected_content"] for a in attachments],
        [a["mime_type"] for a in attachments],
        send_patch,
    )
    # Return empty responses for all rows except the first one.
    empty_responses = []
    for _ in range(len(mock_sheet["rows"]) - 1):
        empty_responses.append(_create_json_response(MOCKED_ROW_NO_ATTACHMENTS))
    _extend_patch(send_patch, empty_responses)

    sheet_with_attachments: Table = authorized_lib.get_sheet(
        123, include="attachments, attachmentFiles", download_path=TEMP_DIR
    )

    send_patch.assert_called()
    assert authorized_lib.current_sheet.selected_includes == [
        "attachments",
        "attachmentFiles",
    ]
    files_row = sheet_with_attachments[:, "attachmentFiles"]
    for row in files_row:
        if row is not None:
            for file, expected_content in zip(row, attachments):
                assert file.exists()
                if expected_content["mime_type"] == "text/plain":
                    try:
                        decoded_content = expected_content["expected_content"].decode()
                    except (UnicodeDecodeError, AttributeError):
                        decoded_content = expected_content["expected_content"]
                    assert file.read_text() == decoded_content
                else:
                    assert file.read_bytes() == expected_content["expected_content"]
    attachment_patches.assert_called()


def test_get_row(lib_with_sheet: Smartsheet, mocker: MockerFixture) -> None:
    patch = _patch_response(lib_with_sheet, mocker, MOCKED_ROW)

    row = lib_with_sheet.get_row(123, native=False)

    patch.assert_called()
    for cell in MOCKED_ROW["cells"]:
        assert cell["value"] == row[cell["columnId"]]


def test_get_native_row(lib_with_sheet: Smartsheet, mocker: MockerFixture) -> None:
    patch = _patch_response(lib_with_sheet, mocker, MOCKED_ROW)

    row = lib_with_sheet.get_row(123, native=True)

    patch.assert_called()
    assert row.id == MOCKED_ROW["id"]
    for cell in MOCKED_ROW["cells"]:
        assert cell["value"] in [c.value for c in row.cells]


@pytest.mark.parametrize(
    "row",
    [
        {NAME_COLUMN_ID: "Mark", ITEM_COLUMN_ID: "Shirt", ZIP_COLUMN_ID: 12345},
        {"Name": "Mark", "Item": "Shirt", "Zip": 12345},
        [
            {"column_id": NAME_COLUMN_ID, "value": "Mark"},
            {"column_id": ITEM_COLUMN_ID, "value": "Shirt"},
            {"column_id": ZIP_COLUMN_ID, "value": 12345},
        ],
        [
            {"title": "Name", "value": "Mark"},
            {"title": "Item", "value": "Shirt"},
            {"title": "Zip", "value": 12345},
        ],
    ],
)
def test_create_row(lib_with_sheet: Smartsheet, row: Union[dict, List]) -> None:
    expected_cells = [
        {"column_id": NAME_COLUMN_ID, "value": "Mark"},
        {"column_id": ITEM_COLUMN_ID, "value": "Shirt"},
        {"column_id": ZIP_COLUMN_ID, "value": 12345},
    ]
    if isinstance(row, dict):
        new_row = lib_with_sheet._create_row_from_dict(row)
    else:
        new_row = lib_with_sheet._create_row_from_list(row)

    for cell, expected_cell in zip(new_row.cells, expected_cells):
        assert cell.column_id == expected_cell["column_id"]
        assert cell.value == expected_cell["value"]


def test_set_row(lib_with_sheet: Smartsheet, mocker: MockerFixture) -> None:
    patch = _patch_response(lib_with_sheet, mocker, MOCKED_UPDATE_ROW_SUCCESS)
    func_patch = _wrap_lib(lib_with_sheet, mocker)

    row = lib_with_sheet.set_row(123, UPDATED_ROW_1, native=True)

    for call in func_patch.Sheets.update_rows.call_args_list:
        sheet_id_called = call[0][0]
        row_called = call[0][1][0]
        assert sheet_id_called == MOCK_SHEET_JSON["id"]
        assert row_called.id == 123
        for cell in row_called.cells:
            assert cell.value in [c["value"] for c in UPDATED_ROW_1]
    patch.assert_called()
    assert row.id == MOCKED_UPDATE_ROW_SUCCESS["result"][0]["id"]
    for cell in UPDATED_ROW_1:
        assert cell["value"] in [c.value for c in row.cells]


def test_set_rows(lib_with_sheet: Smartsheet, mocker: MockerFixture) -> None:
    patch = _patch_response(lib_with_sheet, mocker, MOCKED_UPDATE_ROW_SUCCESS)
    func_patch = _wrap_lib(lib_with_sheet, mocker)

    rows = lib_with_sheet.set_rows(UPDATED_ROWS, native=True)

    for call in func_patch.Sheets.update_rows.call_args_list:
        sheet_id_called = call[0][0]
        rows_called = call[0][1]
        assert sheet_id_called == MOCK_SHEET_JSON["id"]
        for row in rows_called:
            assert row.id in [c.get("rowId", None) for r in UPDATED_ROWS for c in r]
    patch.assert_called()
    assert len(rows) == len(MOCKED_UPDATE_ROW_SUCCESS["result"])


def test_add_rows(lib_with_sheet: Smartsheet, mocker: MockerFixture) -> None:
    patch = _patch_response(lib_with_sheet, mocker, MOCKED_ADD_ROW_SUCCESS)
    func_patch = _wrap_lib(lib_with_sheet, mocker)

    rows = lib_with_sheet.add_rows(NEW_ROWS, native=True)

    for call in func_patch.Sheets.add_rows.call_args_list:
        sheet_id_called = call[0][0]
        rows_called = call[0][1]
        assert sheet_id_called == MOCK_SHEET_JSON["id"]
        for row in rows_called:
            assert row.id is None
    patch.assert_called()
    assert len(rows) == len(MOCKED_ADD_ROW_SUCCESS["result"])


def test_create_sheet(authorized_lib: Smartsheet, mocker: MockerFixture) -> None:
    patch = _patch_response(authorized_lib, mocker, MOCKED_CREATE_SHEET_SUCCESS)
    func_patch = _wrap_lib(authorized_lib, mocker)

    sheet = authorized_lib.create_sheet(NEW_SHEET_NAME, NEW_SHEET_COLUMNS)

    for call in func_patch.Home.create_sheet.call_args_list:
        sheet_called = call[0][0]
        assert sheet_called.name == NEW_SHEET_NAME
        assert sheet_called.columns[0].title == NEW_SHEET_COLUMNS[0]["title"]
        assert sheet_called.columns[0].type == NEW_SHEET_COLUMNS[0]["type"]
    patch.assert_called()
    assert sheet.id == MOCKED_CREATE_SHEET_SUCCESS["result"]["id"]
    assert sheet.name == NEW_SHEET_NAME


def test_search(lib_with_sheet: Smartsheet, mocker: MockerFixture) -> None:
    patch = _patch_response(lib_with_sheet, mocker, MOCKED_SEARCH_RESULTS)
    func_patch = _wrap_lib(lib_with_sheet, mocker)

    results = lib_with_sheet.search(
        "onesie", include="favoriteFlag", scopes="comments,cellData"
    )

    for call in func_patch.Search.search.call_args_list:
        query_called = call[0][0]
        include_called = call[0][1]
        scope_called = call[0][4]
        assert query_called == "onesie"
        assert include_called == ["favoriteFlag"]
        assert scope_called == ["comments", "cellData"]
    patch.assert_called()
    assert len(results) == len(MOCKED_SEARCH_RESULTS["results"])
    for result in results:
        assert result.object_id in [
            r["objectId"] for r in MOCKED_SEARCH_RESULTS["results"]
        ]


def test_get_row_from_number(lib_with_sheet: Smartsheet, mocker: MockerFixture) -> None:
    row_id = lib_with_sheet._get_row_from_number(1)

    assert row_id == MOCK_SHEET_JSON["rows"][0]["id"]



import time
from datetime import datetime
from pathlib import Path
from typing import Any, List, Optional, Union

from RPA.application import BaseApplication, COMError
from RPA.Email.common import counter_duplicate_path

RECIPIENT_TYPE_TO = 1
RECIPIENT_TYPE_CC = 2
RECIPIENT_TYPE_BCC = 3


class InvalidNamesError(Exception):
    """Raised when email recipient names are not recognized."""


class Application(BaseApplication):
    # pylint: disable=C0301
    """`Outlook.Application` is a library for controlling the Outlook application.

    **About Email Filtering**

    Emails can be filtered according to specification set by Restrict method of the
    Item class https://docs.microsoft.com/en-us/office/vba/api/outlook.items.restrict.

    Couple of examples:

    .. code-block:: robotframework

        Get Emails
        ...   email_filter=[Subject]='test email'

        Move Emails
        ...   email_filter=[SenderEmailAddress]='hello@gmail.com'

    **Examples**

    **Robot Framework**

    .. code-block:: robotframework

        *** Settings ***
        Library                 RPA.Outlook.Application
        Task Setup              Open Application
        Suite Teardown          Quit Application

        *** Variables ***
        ${RECIPIENT}            address@domain.com

        *** Tasks ***
        Send email
            Send Email         recipients=${RECIPIENT}
            ...                subject=This is the subject
            ...                body=This is the message body
            ...                attachments=approved.png

    **Python**

    .. code-block:: python

        from RPA.Outlook.Application import Application

        def send_email():
            app = Application()
            app.open_application()
            app.send_email(
                recipients='EMAILADDRESS_1, EMAILADDRESS_2',
                subject='email subject',
                body='email body message',
                attachments='../orders.csv'

    For more information, see: https://docs.microsoft.com/en-us/previous-versions/office/developer/office-2007/bb219950(v=office.12)
    """  # noqa: E501

    APP_DISPATCH = "Outlook.Application"

    def send_email(
        self,
        recipients: Union[str, List[str]],
        subject: str,
        body: str,
        html_body: bool = False,
        attachments: Optional[Union[str, List[str]]] = None,
        save_as_draft: bool = False,
        cc_recipients: Optional[Union[str, List[str]]] = None,
        bcc_recipients: Optional[Union[str, List[str]]] = None,
        reply_to: Optional[Union[str, List[str]]] = None,
        check_names: bool = False,
    ) -> bool:
        """Send email with Outlook

        :param recipients: list of addresses
        :param subject: email subject
        :param body: email body
        :param html_body: True if body contains HTML, defaults to False
        :param attachments: list of filepaths to include in the email, defaults to []
        :param save_as_draft: email is saved as draft when `True`
        :param cc_recipients: list of addresses for CC field, default None
        :param bcc_recipients: list of addresses for BCC field, default None
        :param reply_to: list of addresses for changing email's reply-to field,
         default None
        :param check_names: all recipients are checked if the email address is
         recognized on `True`, default `False`
        :return: `True` if there were no errors

        Example:

        .. code-block:: python

            library = Outlook()
            library.open_application()
            cc_recipients = ["recipient3@domain.com","recipient4@domain.com"]
            library.send_email(
                recipients="recipient1@domain.com",
                cc_recipients=cc_recipients,
                bcc_recipients="recipient3@domain.com;recipient4@domain.com",
                subject="hello from Outlook",
                body="empty body",
                attachments=os.path.join(os.path.curdir, "example.xslx")
            )

        .. code-block:: robotframework

            ${cc}=  Create List   recipient3@domain.com   recipient4@domain.com
            Send Email
            ...    recipients=recipient1@domain.com
            ...    cc_repients=${cc}
            ...    bcc_repients=recipient5@domain.com;recipient6@domain.com
            ...    subject=hello from Outlook
            ...    body=empty body
            ...    attachments=${CURDIR}${/}example.xlsx
        """
        # pylint: disable=no-member
        attachments = attachments or []
        if not isinstance(attachments, list):
            attachments = str(attachments).split(";")

        try:
            mail = self.app.CreateItem(0)
            mail.Subject = subject
            self._add_all_recipients(
                mail, recipients, cc_recipients, bcc_recipients, reply_to, check_names
            )

            if html_body:
                mail.HTMLBody = body
            else:
                mail.Body = body

            self._add_attachments(mail, attachments)

            if save_as_draft:
                mail.Save()
                self.logger.debug("Email draft saved")
            else:
                mail.Send()
                self.logger.debug("Email sent")
        # On non-Windows OS `COMError` is `Exception`.
        except COMError as exc:  # pylint: disable=broad-except
            self.logger.error(
                f"Mail {'saving' if save_as_draft else 'sending'} failed: %s", exc
            )
            return False
        return True

    def _add_all_recipients(
        self, email, recipients, cc_recipients, bcc_recipients, reply_to, check_names
    ):
        invalid_names = []

        self._recipients_to_email(
            email, invalid_names, check_names, recipients, RECIPIENT_TYPE_TO
        )
        self._recipients_to_email(
            email, invalid_names, check_names, cc_recipients, RECIPIENT_TYPE_CC
        )
        self._recipients_to_email(
            email, invalid_names, check_names, bcc_recipients, RECIPIENT_TYPE_BCC
        )

        if len(invalid_names) > 0:
            raise InvalidNamesError(
                f"Failed to resolve recipient(s): {';'.join(invalid_names)}"
            )

        reply_to = self._recipients_to_list(reply_to)
        for recipient in reply_to:
            email.ReplyRecipients.Add(recipient)

    def _recipients_to_email(
        self, email, invalid_names, check_names, recipients, recipient_type
    ):
        recipients = self._recipients_to_list(recipients)
        for recipient in recipients:
            rec = email.Recipients.Add(recipient)
            # 1 = To, 2 = CC, 3 = BCC
            rec.Type = recipient_type
            if check_names:
                result = rec.Resolve()  # Attempt to resolve each recipient
                if not result:
                    invalid_names.append(recipient)

    def _recipients_to_list(self, recipients):
        recipients = recipients or []
        recipients = (
            recipients if isinstance(recipients, list) else recipients.split(";")
        )
        return recipients

    def _add_attachments(self, email, attachments):
        for attachment in attachments:
            filepath = Path(attachment).absolute()
            email.Attachments.Add(str(filepath))

    def _is_email_too_old(self, email):
        now = datetime.now().timestamp()
        msg_time = email["ReceivedTimestamp"]
        timediff = now - msg_time
        return timediff > 30.0

    def _check_for_matching(self, criterion, email):
        crit_key, crit_val = criterion.split(":", 1)
        crit_key = crit_key.upper()
        crit_val = crit_val.lower()
        match_found = None

        if crit_key.upper() == "SUBJECT" and crit_val in email["Subject"].lower():
            match_found = email
        elif crit_key.upper() == "SENDER" and crit_val in email["Sender"].lower():
            match_found = email
        elif crit_key.upper() == "BODY" and crit_val in email["Body"].lower():
            match_found = email

        return match_found

    def wait_for_email(
        self, criterion: str = None, timeout: float = 5.0, interval: float = 1.0
    ) -> Any:
        """Wait for email matching `criterion` to arrive into mailbox.

        :param criterion: email filter to wait for, defaults to ""
        :param timeout: total time in seconds to wait for email, defaults to 5.0
        :param interval: time in seconds for new check, defaults to 1.0
        :return: list of messages or False

        Possible wait criterias are: SUBJECT, SENDER and BODY

        Example:

        .. code-block:: robotframework

            Wait for Email     SUBJECT:rpa task calling    timeout=300    interval=10
        """
        if criterion is None:
            self.logger.warning(
                "Wait for message requires criteria for which message to wait for."
            )
            return False
        end_time = time.time() + float(timeout)
        namespace = self.app.GetNamespace("MAPI")
        inbox = namespace.GetDefaultFolder(6)
        while time.time() < end_time:
            messages = inbox.Items
            messages.Sort("[ReceivedTime]", True)
            for msg in messages:
                if type(msg).__name__ != "_MailItem":
                    continue
                m = self._mail_item_to_dict(msg)
                if self._is_email_too_old(m):
                    break
                else:
                    match_found = self._check_for_matching(criterion, m)
                    if match_found:
                        return match_found
            time.sleep(interval)
        raise AssertionError("Did not find matching message in the Outlook inbox")

    def get_emails(
        self,
        account_name: str = None,
        folder_name: str = None,
        email_filter: str = None,
        save_attachments: bool = False,
        attachment_folder: str = None,
        sort: bool = False,
        sort_key: str = None,
        sort_descending: bool = True,
    ) -> List:
        """Get emails from a specified email folder. Can be used to save attachments.

        :param account_name: needs to be given if there are shared accounts in use,
         defaults to None
        :param folder_name: target folder where to get emails from, default Inbox
        :param email_filter: how to filter email, default no filter,
         ie. all emails in folder
        :param save_attachments: if attachments should be saved, defaults to False
        :param attachment_folder: target folder where attachments are saved,
         defaults to current directory
        :param sort: if emails should be sorted, defaults to False
        :param sort_key: needs to be given if emails are to be sorted
        :param sort_descending: set to False for ascending sort, defaults to True
        :return: list of emails (list of dictionaries)

        Example:

        .. code-block:: robotframework

            ${emails}=  Get Emails
            ...    email_folder=priority
            ...    email_filter=[Subject]='incoming order'
            ...    save_attachments=True
            ...    attachment_folder=%{ROBOT_ROOT}${/}attachments
            ...    sort=True
            ...    sort_key=Received
            ...    sort_descending=False
        """
        folder = self._get_folder(account_name, folder_name)
        folder_messages = folder.Items if folder else []
        messages = []
        if folder_messages and email_filter:
            try:
                folder_messages = folder_messages.Restrict(email_filter)
            except Exception:  # pylint: disable=broad-except
                raise AttributeError(  # pylint: disable=raise-missing-from
                    "Invalid email filter '%s'" % email_filter
                )
        if folder_messages and sort:
            sort_key = sort_key or "ReceivedTime"
            try:
                folder_messages.Sort(f"[{sort_key}]", sort_descending)
            except Exception:  # pylint: disable=broad-except
                raise AttributeError(  # pylint: disable=raise-missing-from
                    "Invalid email sort key '%s'" % sort_key
                )
        for message in folder_messages:
            if type(message).__name__ != "_MailItem":
                continue
            if save_attachments:
                self.save_email_attachments(message.Attachments, attachment_folder)
            messages.append(self._mail_item_to_dict(message))
        return messages

    def _get_folder(self, account_name, email_folder):
        namespace = self.app.GetNamespace("MAPI")
        if not account_name and not email_folder:
            self.logger.warning("Getting items from default account inbox")
            return namespace.GetDefaultFolder(6)
        email_folder = email_folder or "Inbox"
        if account_name:
            account_folder = self._get_account_folder(namespace, account_name)
            if account_folder:
                folder = self._get_matching_folder(email_folder, account_folder)
            else:
                raise AttributeError("Did not find account by name '%s'" % account_name)
        else:
            folder = self._get_matching_folder(email_folder, None)
        return folder if folder else []

    def _get_account_folder(self, namespace, account_name):
        account_folder = None
        for f in namespace.Folders:
            if f.Name == account_name:
                account_folder = f
                break
        return account_folder

    def _get_matching_folder(self, folder_name, folder=None):
        if not folder or isinstance(folder, str):
            folders = self.app.GetNamespace("MAPI").Folders
        elif isinstance(folder, list):
            folders = folder
        else:
            folders = folder.Folders
        for f in folders:
            if folder_name == f.Name:
                self.logger.debug("Found matching folder: %s", f.Name)
                return f
            emails = self._get_matching_folder(folder_name, f)
            if emails:
                return emails
        return None

    def save_email_attachments(
        self, attachments: Any, attachment_folder: str, overwrite: bool = False
    ) -> None:
        """Save email attachments.

        Note. Keyword "Get Emails" can be also used to save attachments.

        :param attachments: all attachments from email or single attachment
        :param attachment_folder: target folder where attachments are saved,
            defaults to current directory
        :param overwrite: overwrite existing file if True, defaults to False

        Example:

        .. code-block:: robotframework

            ${emails} =  Get Emails
            ...    email_folder=priority
            FOR  ${email}  IN   @{emails}
                FOR  ${attachment}  IN  @{email}[Attachments]
                    IF  ${attachment}[size] < 100000   # bytes
                        Save Email Attachments
                        ...  ${attachment}
                        ...  ${CURDIR}${/}attachments
                    ELSE IF  ".pdf" in "${attachment}[filename]"
                        Save Email Attachments
                        ...  ${attachment}
                        ...  ${CURDIR}${/}attachments${/}pdf
                    END
                END
            END
        """
        attachment_target = Path(attachment_folder) if attachment_folder else Path(".")
        if isinstance(attachments, dict):
            email_attachments = [attachments["item"]]
        else:
            email_attachments = attachments
        for attachment in email_attachments:
            file_path = (attachment_target / attachment.FileName).absolute()
            if not overwrite:
                file_path = counter_duplicate_path(file_path)
            attachment.SaveAsFile(file_path)

    def mark_email_as_read(self, email: Any, read: bool = True) -> None:
        """Mark email 'read' property. Can be used to mark email as unread.

        :param email: target email
        :param read: True marks email as Read, False as Unread

        Example:

        .. code-block:: robotframework

            ${emails}=  Get Emails
            # Mark all as read
            FOR  ${email}  IN  @{emails}
                Mark Email As Read  ${email}
            END

            # Mark all as unread
            FOR  ${email}  IN  @{emails}
                Mark Email As Read  ${email}  False
            END
        """
        read_value = not read
        if type(email).__name__ == "_MailItem":
            email.UnRead = read_value
            email.Save()
        elif isinstance(email, dict):
            email["object"].UnRead = read_value
            email["object"].Save()

    def move_emails(
        self,
        account_name: str = None,
        source_folder: str = None,
        email_filter: Any = None,
        target_folder: str = None,
        mark_as_read: bool = True,
    ) -> bool:
        """Move emails from source folder to target folder.

        Use of "account_name" is recommended if there are shared accounts in use.

        :param account_name: needs to be given if there are shared accounts in use,
         defaults to None
        :param source_folder: folder where source emails exist
        :param email_filter: how to filter email, default no filter,
         ie. all emails in folder
        :param target_folder: folder where emails are moved into
        :param mark_as_read: mark emails as read after move, defaults to True
        :return: True if move operation was success, False if not

        Python example.

        .. code-block:: python

            outlook = RPA.Outlook.Application()

            # moving messages from Inbox to target_folder
            outlook.move_emails(
                target_folder='Processed Invoices',
                email_filter="[Subject]='incoming invoice'"
            )

            # moving messages from source_folder to target_folder
            outlook.move_emails(
                source_folder='Incoming Invoices',
                target_folder='Processed Invoices',
                email_filter="[Subject]='incoming invoice'"
            )

            # move message objects from `get_emails` result
            emails = outlook.get_emails("[Subject]='incoming invoice'")
            outlook.move_emails(
                target_folder='Processed Invoices',
                email_filter=emails
            )

        Robot Framework example.

        .. code-block:: robotframework

            # moving messages from Inbox to target_folder
            Move Emails
            ...    target_folder=Processed Invoices
            ...    email_filter=[Subject]='incoming invoice'

            # moving messages from source_folder to target_folder
            Move Emails
            ...    source_folder=Incoming Invoices
            ...    target_folder=Processed Invoices
            ...    email_filter=[Subject]='incoming invoice'

            # moving message objects from `Get Emails` result
            ${emails}=    Get Emails    [Subject]='incoming invoice'
            Move Emails
            ...    target_folder=Processed Invoices
            ...    email_filter=${emails}
        """
        if not target_folder:
            raise AttributeError("Can't move emails without target_folder")
        tf = self._get_folder(account_name, target_folder)
        if not tf or tf.Name != target_folder:
            raise AttributeError("Did not find target folder")
        self.logger.info("Found target folder: %s", tf.Name)
        if email_filter and not isinstance(email_filter, str):
            folder_messages = (
                email_filter if isinstance(email_filter, list) else [email_filter]
            )
        else:
            folder = self._get_folder(account_name, source_folder)
            folder_messages = folder.Items if folder else []
            try:
                folder_messages = folder_messages.Restrict(email_filter)
            except Exception:  # pylint: disable=broad-except
                raise AttributeError(  # pylint: disable=raise-missing-from
                    "Invalid email filter '%s'" % email_filter
                )
        if not folder_messages or len(folder_messages) == 0:
            self.logger.warning("Did not find emails to move")
            return False
        self.logger.info("Found %d emails to move", len(folder_messages))
        for m in folder_messages:
            m = m["object"] if isinstance(m, dict) else m
            if mark_as_read:
                m.UnRead = False
            m.Move(tf)
        return True

    def _mail_item_to_dict(self, mail_item):
        mi = mail_item
        response = {
            "Sender": self._get_sender_email_address(mi),
            "To": [],
            "CC": [],
            "BCC": [],
            "Subject": mi.Subject,
            "Body": mi.Body,
            "Attachments": [
                {"filename": a.FileName, "size": a.Size, "item": a}
                for a in mi.Attachments
            ],
            "Size": mi.Size,
            "object": mi,
        }
        rt = getattr(mail_item, "ReceivedTime", "<UNKNOWN>")
        response["ReceivedTime"] = rt.isoformat() if rt != "<UNKNOWN>" else rt
        response["ReceivedTimestamp"] = (
            datetime(
                rt.year, rt.month, rt.day, rt.hour, rt.minute, rt.second
            ).timestamp()
            if rt
            else None
        )
        so = getattr(mail_item, "SentOn", "<UNKNOWN>")
        response["SentOn"] = so.isoformat() if so != "<UNKNOWN>" else so
        self._handle_recipients(mi.Recipients, response)
        return response

    def _get_recipient_email_address(self, recipient):
        email_address = None
        try:
            email_address = recipient.AddressEntry.GetExchangeUser().PrimarySmtpAddress
        except Exception:  # pylint: disable=broad-except
            email_address = recipient.Address
        return email_address

    def _handle_recipients(self, recipients, response):
        for r in recipients:
            if r.Type == 1:
                response["To"].append(
                    {
                        "name": r.Name,
                        "email": self._get_recipient_email_address(r),
                    }
                )
            elif r.Type == 2:
                response["CC"].append(
                    {
                        "name": r.Name,
                        "email": self._get_recipient_email_address(r),
                    }
                )
            elif r.Type == 3:
                response["BCC"].append(
                    {
                        "name": r.Name,
                        "email": self._get_recipient_email_address(r),
                    }
                )

    def _get_sender_email_address(self, mail_item):
        mi = mail_item
        return (
            mi.Sender.GetExchangeUser().PrimarySmtpAddress
            if mi.SenderEmailType == "EX"
            else mi.SenderEmailAddress
        )

    def _address_entry_to_dict(self, address_entry):
        ae = address_entry
        return {"Name": ae.Name, "Address": ae.Address}



import re
import urllib.parse

import pytest
from selenium.webdriver import ChromeOptions

from RPA.Browser.Selenium import Selenium, ensure_scheme

from . import RESOURCES_DIR, temp_filename


CHROMIUM_HEADLESS = "--headless=new"


@pytest.fixture
def library():
    lib = Selenium()
    yield lib
    lib.close_all_browsers()


def get_chrome_options():
    """Returns `ChromeOptions` with custom arguments and capabilities."""
    options = ChromeOptions()
    options.add_argument(CHROMIUM_HEADLESS)
    options.set_capability("acceptInsecureCerts", True)
    return options


class TestSelenium:
    """`RPA.Browser.Selenium` library tests."""

    def test_print_to_pdf(self, library):
        testfile = RESOURCES_DIR / "browser_docs.html"
        library.open_available_browser(
            f"file://{testfile}", headless=True, browser_selection="Chrome"
        )
        with temp_filename(suffix=".pdf") as tmp_file:
            library.print_to_pdf(tmp_file)
            with open(tmp_file, "rb") as stream:
                data = stream.read()
            assert b"browser_docs.html" or b"selenium" in data

    @pytest.mark.xfail(reason="Firefox not available")
    def test_print_to_pdf_exception_on_non_supported_driver(self, library):
        testfile = RESOURCES_DIR / "browser_docs.html"
        library.open_available_browser(
            f"file://{testfile}", browser_selection="Firefox", headless=True
        )
        err_expr = re.compile(r"PDF printing works only with Chromium-based browsers")
        with pytest.raises(NotImplementedError, match=err_expr):
            library.print_to_pdf()

    @pytest.mark.parametrize(
        "options",
        [
            {
                "arguments": [CHROMIUM_HEADLESS],
                "capabilities": {"acceptInsecureCerts": True},
            },
            {
                "arguments": CHROMIUM_HEADLESS,
                "capabilities": "acceptInsecureCerts:True",
            },
            f"add_argument('{CHROMIUM_HEADLESS}');set_capability('acceptInsecureCerts', True)",
            get_chrome_options(),
        ],
    )
    def test_options_normalization(self, library, options):
        options_obj = library.normalize_options(options, browser="Chrome")
        args = options_obj.to_capabilities()["goog:chromeOptions"]["args"]
        assert options_obj.accept_insecure_certs
        assert CHROMIUM_HEADLESS in args

    def test_unrecognized_option(self, library):
        options = {"argument": CHROMIUM_HEADLESS}  # expecting "arguments" instead
        with pytest.raises(TypeError):
            library.normalize_options(options, browser="Chrome")

    def test_custom_options(self, library):
        path = "path/to/chrome"
        options = {"binary_location": path}
        options_obj = library.normalize_options(options, browser="Chrome")
        assert options_obj.binary_location == path


@pytest.mark.parametrize(
    "url,default,scheme",
    [
        ("https://www.google.com", "https", "https"),
        ("http://www.google.com", "https", "http"),
        ("www.google.com", "https", "https"),
        ("about:config", "https", "about"),
        ("www.google.com", None, ""),
    ],
)
def test_ensure_scheme(url, default, scheme):
    result = ensure_scheme(url, default)
    parsed = urllib.parse.urlsplit(result)
    assert parsed.scheme == scheme
